# Generated by Django 3.1.2 on 2020-10-18 16:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0046_auto_20201018_1743'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='typ',
            field=models.CharField(blank=True, choices=[('', '-'), ('kunde_registriert', 'Registrierungsmail Kunde'), ('bestellung_rechnung', 'Rechnung'), ('bestellung_stock_warning', 'Warnung zum Lagerbestand')], default='', max_length=50, verbose_name='Typ'),
        ),
    ]



from django.urls import path

from kmuhelper.modules.emails import views

#######################

urlpatterns = [
    path('<int:object_id>/view/', views.email_view,
         name="email-view"),
    path('attachments/<int:object_id>/view/', views.attachment_view,
         name="attachment-view"),

    path('', views.email_index,
         name="email-index"),
]



# Generated by Django 4.1 on 2023-01-27 17:34

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("kmuhelper", "0080_auto_20220311_2125"),
    ]

    operations = [
        migrations.AddField(
            model_name="zahlungsempfaenger",
            name="iban",
            field=models.CharField(
                blank=True,
                default="",
                help_text="IBAN mit Leerzeichen (Nur verwendet im Modus 'Ohne Referenz')",
                max_length=26,
                validators=[
                    django.core.validators.RegexValidator(
                        "^(CH|LI)[0-9]{2}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{1}$",
                        "Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x",
                    )
                ],
                verbose_name="IBAN",
            ),
        ),
        migrations.AddField(
            model_name="zahlungsempfaenger",
            name="mode",
            field=models.CharField(
                choices=[("QRR", "QR-Referenz"), ("NON", "Ohne Referenz")],
                default="QRR",
                max_length=15,
                verbose_name="Modus",
            ),
        ),
        migrations.AlterField(
            model_name="zahlungsempfaenger",
            name="firmenname",
            field=models.CharField(
                help_text="Name der Firma oder des Empfängers",
                max_length=70,
                verbose_name="Firmennname",
            ),
        ),
        migrations.AlterField(
            model_name="zahlungsempfaenger",
            name="firmenuid",
            field=models.CharField(
                blank=True,
                default="",
                help_text="UID der Firma - Format: CHE-123.456.789 (Mehrwertsteuernummer)",
                max_length=15,
                validators=[
                    django.core.validators.RegexValidator(
                        "^CHE-[0-9]{3}\\.[0-9]{3}\\.[0-9]{3}$",
                        "Bite benutze folgendes Format: CHE-123.456.789",
                    )
                ],
                verbose_name="Firmen-UID",
            ),
        ),
        migrations.AlterField(
            model_name="zahlungsempfaenger",
            name="logourl",
            field=models.URLField(
                blank=True,
                default="",
                help_text="URL eines Bildes (.jpg/.png) - Wird auf die Rechnung gedruckt.",
                validators=[
                    django.core.validators.RegexValidator(
                        "^[0-9a-zA-Z\\-\\.\\|\\?\\(\\)\\*\\+&\"'_:;/]+\\.(png|jpg)$",
                        "Nur folgende Zeichen gestattet: 0-9a-zA-Z-_.:;/|?&()\"'*+ - Muss auf .jpg/.png enden.",
                    )
                ],
                verbose_name="Logo (URL)",
            ),
        ),
        migrations.AlterField(
            model_name="zahlungsempfaenger",
            name="qriban",
            field=models.CharField(
                blank=True,
                default="",
                help_text="QR-IBAN mit Leerzeichen (Nur verwendet im Modus 'QR-Referenz')",
                max_length=26,
                validators=[
                    django.core.validators.RegexValidator(
                        "^(CH|LI)[0-9]{2}\\s3[0-9]{3}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{4}\\s[0-9]{1}$",
                        "Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x",
                    )
                ],
                verbose_name="QR-IBAN",
            ),
        ),
    ]



# Generated by Django 3.0.4 on 2020-06-27 14:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0015_bestellung_fix_summe'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bestellung',
            name='rechnungsadresse_adresszeile2',
            field=models.CharField(blank=True, default='', help_text='Wird in QR-Rechnung NICHT verwendet!', max_length=50, verbose_name='Adresszeile 2'),
        ),
        migrations.AlterField(
            model_name='kunde',
            name='rechnungsadresse_adresszeile2',
            field=models.CharField(blank=True, default='', help_text='Wird in QR-Rechnung NICHT verwendet!', max_length=50, verbose_name='Adresszeile 2'),
        ),
    ]



# Generated by Django 3.1.3 on 2021-02-02 20:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0047_auto_20201018_1851'),
    ]

    operations = [
        migrations.AddField(
            model_name='email',
            name='notes',
            field=models.TextField(blank=True, default='', verbose_name='Notizen'),
        ),
        migrations.AlterField(
            model_name='email',
            name='html_template',
            field=models.CharField(max_length=100, verbose_name='Dateiname der Vorlage'),
        ),
    ]



# Generated by Django 3.1.3 on 2021-02-06 15:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0049_email_data'),
    ]

    operations = [
        migrations.AddField(
            model_name='produkt',
            name='lieferant_url',
            field=models.URLField(blank=True, default='', verbose_name='Lieferantenurl (Für Nachbestellungen)'),
        ),
    ]



# Generated by Django 3.2 on 2021-04-12 14:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0056_auto_20210409_2340'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='email',
            options={'default_permissions': ('add', 'change', 'view', 'delete', 'send'), 'verbose_name': 'E-Mail', 'verbose_name_plural': 'E-Mails'},
        ),
        migrations.AddField(
            model_name='email',
            name='sent',
            field=models.BooleanField(default=False, verbose_name='Gesendet?'),
        ),
    ]



"""Useful constants for api views"""

from django.http import JsonResponse

APIKEY_INVALID = JsonResponse({
    "error": "apikey-invalid",
    "message": "Your api key is invalid!"
}, status=403)

ENDPOINT_NOT_FOUND = JsonResponse({
    "error": "endpoint-not-found",
    "message": "This endpoint does not exist!"
}, status=404)

NO_PERMISSION_APIKEY = JsonResponse({
    "error": "no-permission",
    "message": "You don't have permission to access this endpoint! (authentication method: api key)"
}, status=403)

NO_PERMISSION_SESSION = JsonResponse({
    "error": "no-permission",
    "message": "You don't have permission to access this endpoint! (authentication method: session)"
}, status=403)

NOT_AUTHENTICATED = JsonResponse({
    "error": "not-authenticated",
    "message": "You must authenticate yourself to use this endpoint!"
}, status=401)

OBJ_NOT_FOUND = JsonResponse({
    "error": "object-not-found",
    "message": "The object couln't be found!",
}, status=400)

SUCCESSFULLY_CHANGED = JsonResponse({
    "success": "change-successfull",
    "message": "Successfully changed object!"
}, status=200)



# Generated by Django 3.0.4 on 2020-07-15 14:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0017_todoentry'),
    ]

    operations = [
        migrations.AlterField(
            model_name='todoentry',
            name='beschrieb',
            field=models.TextField(blank=True, default='', verbose_name='Beschrieb'),
        ),
        migrations.AlterField(
            model_name='todoentry',
            name='priority',
            field=models.IntegerField(blank=True, default=0, verbose_name='Priorität'),
        ),
    ]



# Generated by Django 3.0.4 on 2020-07-16 12:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0019_auto_20200715_1931'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='kosten',
            name='versteuerbar',
        ),
        migrations.AddField(
            model_name='kosten',
            name='mwstsatz',
            field=models.FloatField(choices=[(0.0, '0.0% (Mehrwertsteuerfrei)'), (7.7, '7.7% (Normalsatz)'), (3.7, '3.7% (Sondersatz für Beherbergungsdienstleistungen)'), (2.5, '2.5% (Reduzierter Satz)')], default=7.7, verbose_name='Mehrwertsteuersatz'),
        ),
        migrations.AlterField(
            model_name='produkt',
            name='mwstsatz',
            field=models.FloatField(choices=[(0.0, '0.0% (Mehrwertsteuerfrei)'), (7.7, '7.7% (Normalsatz)'), (3.7, '3.7% (Sondersatz für Beherbergungsdienstleistungen)'), (2.5, '2.5% (Reduzierter Satz)')], default=7.7, verbose_name='Mehrwertsteuersatz'),
        ),
    ]



"""PDF creator for invoices and delivery notes"""

from datetime import datetime

from kmuhelper import settings
from kmuhelper.translations import autotranslate_mengenbezeichnung, autotranslate_kosten_name, langselect
from kmuhelper.utils import formatprice
from kmuhelper.modules.pdfgeneration.base import PDFGenerator

from reportlab.platypus import Table, TableStyle, Paragraph, Spacer, TopPadder, Flowable
from reportlab.lib.utils import ImageReader
from reportlab.lib.units import mm
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.colors import black
from reportlab.graphics.shapes import Drawing
from reportlab.graphics.barcode import qr
from reportlab.graphics import renderPDF

from django.utils import translation
_ = translation.gettext

style_default = ParagraphStyle("Normal", fontname="Helvetica")
style_bold = ParagraphStyle("Bold", fontname="Helvetica-Bold")

#####


class _PDFOrderPriceTable(Table):
    COLWIDTHS = [26*mm, 80*mm, 20*mm, 20*mm, 20*mm, 20*mm]

    @classmethod
    def from_order(cls, order, lang="de", show_payment_conditions=None):
        data = [(_("Art-Nr."), _("Bezeichnung"), _("Anzahl"),
                 _("Einheit"), _("Preis"), _("Total"))]

        # Products

        h_products = 0

        for bp in order.produkte.through.objects.filter(bestellung=order):
            zwsumohnerabatt = bp.zwischensumme_ohne_rabatt()
            data.append((
                bp.produkt.artikelnummer,
                Paragraph(langselect(bp.produkt.name, lang)),
                str(bp.menge),
                langselect(autotranslate_mengenbezeichnung(bp.produkt.mengenbezeichnung), lang),
                formatprice(bp.produktpreis),
                formatprice(zwsumohnerabatt)
            ))
            h_products += 1
            if bp.rabatt:
                data.append((
                    "",
                    "- "+_("Rabatt"),
                    str(bp.rabatt),
                    "%",
                    formatprice(zwsumohnerabatt),
                    formatprice(bp.nur_rabatt())
                ))
                h_products += 1
            if bp.bemerkung:
                data.append((
                    "",
                    Paragraph(f"- <b>{bp.bemerkung}</b>"),
                    "",
                    "",
                    "",
                    ""
                ))
                h_products += 1

        # Costs

        h_costs = 0

        for bk in order.kosten.through.objects.filter(bestellung=order):
            data.append((
                "",
                Paragraph(langselect(autotranslate_kosten_name(bk.name), lang)),
                "",
                "",
                "",
                formatprice(bk.preis)
            ))
            h_costs += 1
            if bk.rabatt:
                data.append((
                    "",
                    "- "+_("Rabatt"),
                    str(bk.rabatt),
                    "%",
                    formatprice(bk.zwischensumme_ohne_rabatt()),
                    formatprice(bk.nur_rabatt())
                ))
                h_costs += 1
            if bk.bemerkung:
                data.append((
                    "",
                    Paragraph(f"- <b>{bk.bemerkung}</b>"),
                    "",
                    "",
                    "",
                    ""
                ))
                h_costs += 1

        # VAT

        h_vat = 0

        mwstdict = dict(order.mwstdict())
        for mwstsatz in mwstdict:
            data.append((
                "",
                _("MwSt"),
                mwstsatz,
                "%",
                formatprice(float(mwstdict[mwstsatz])),
                formatprice(float(mwstdict[mwstsatz]*(float(mwstsatz)/100)))
            ))
            h_vat += 1

        # Total & Payment conditions

        if show_payment_conditions is None:
            show_payment_conditions = settings.get_db_setting(
                "print-payment-conditions", False)

        if show_payment_conditions:
            payconds = order.get_paymentconditions()
            totaltext = _("Rechnungsbetrag, zahlbar netto innert %s Tagen") % payconds[-1]["days"]
        else:
            totaltext = _("Rechnungsbetrag")

        data.append((
            Paragraph(f"<b>{totaltext}</b>"),
            "",
            "",
            "CHF",
            "",
            formatprice(order.fix_summe),
        ))

        h_paycond = 0
        if show_payment_conditions:
            for paycond in payconds:
                if paycond["percent"] != 0.0:
                    data.append((
                        _("%(days)s Tage %(percent)s%% Skonto") % {"days": paycond["days"],
                                                                  "percent": paycond["percent"]},
                        "",
                        "",
                        "CHF",
                        "",
                        formatprice(paycond["price"]),
                    ))
                    h_paycond += 1

        # Style

        style = [
            # Horizontal lines
            # Header
            ('LINEABOVE', (0, 0), (-1, 0), 1, black),
            # Header/Products divider
            ('LINEBELOW', (0, 0), (-1, 0), 1, black),
            # Products/Costs divider
            ('LINEBELOW', (0, h_products), (-1, h_products), 0.5, black),
            # Costs/VAT divider
            ('LINEBELOW', (0, h_products+h_costs), (-1, h_products+h_costs), 0.5, black),
            # VAT/Footer divider
            ('LINEBELOW', (0, h_products+h_costs+h_vat), (-1, h_products+h_costs+h_vat), 1, black),
            # Footer
            ('LINEBELOW', (0, -1), (-1, -1), 1, black),

            # Span for total line
            ('SPAN', (0, -1-h_paycond), (2, -1-h_paycond)),

            # Horizontal alignment (same for all rows)
            ('ALIGN', (-1, 0), (-1, -1), "RIGHT"),
            ('ALIGN', (-2, 0), (-2, -1), "RIGHT"),
            ('ALIGN', (-4, 0), (-4, -1), "RIGHT"),
            # Vertical alignment (same for whole table)
            ('VALIGN', (0, 0), (-1, -1), "TOP"),
            # Bold total line
            ('FONTNAME', (0, -1-h_paycond), (-1, -1-h_paycond), "Helvetica-Bold"),
        ]

        return cls(data, repeatRows=1, style=TableStyle(style), colWidths=cls.COLWIDTHS)


class _PDFOrderProductTable(Table):
    COLWIDTHS = [36*mm, 110*mm, 20*mm, 20*mm]

    @classmethod
    def from_order(cls, order, lang="de"):
        data = [(_("Art-Nr."), _("Bezeichnung"), _("Anzahl"), _("Einheit"))]

        productcount = 0

        # Products

        for bp in order.produkte.through.objects.filter(bestellung=order):
            data.append((
                bp.produkt.artikelnummer,
                Paragraph(langselect(bp.produkt.name, lang)),
                str(bp.menge),
                langselect(autotranslate_mengenbezeichnung(bp.produkt.mengenbezeichnung), lang),
            ))
            if bp.bemerkung:
                data.append((
                    "",
                    Paragraph(f"- <b>{bp.bemerkung}</b>"),
                    "",
                    ""
                ))

            productcount += bp.menge

        # Total

        data.append((
            _("Anzahl Produkte"),
            "",
            str(productcount),
            ""
        ))

        style = TableStyle([
            ('LINEABOVE', (0, 0), (-1, 0), 1, black),
            ('LINEBELOW', (0, 0), (-1, 0), 1, black),
            ('LINEABOVE', (0, -1), (-1, -1), 1, black),
            ('LINEBELOW', (0, -1), (-1, -1), 1, black),
            ('ALIGN', (1, -1), (1, -1), "CENTER"),
            ('FONTNAME', (0, -1), (-1, -1), "Helvetica-Bold"),
            ('VALIGN', (0, 0), (-1, -1), "TOP"),
        ])

        return cls(data, repeatRows=1, style=style, colWidths=cls.COLWIDTHS)


class _PDFOrderQrInvoice(Flowable):
    @classmethod
    def from_order(cls, order, add_cut_lines=True):
        elem = cls()
        elem.width = 210
        elem.height = 110
        elem._fixedWidth = 210
        elem._fixedHeight = 110
        elem.add_cut_info = add_cut_lines
        elem.order = order
        return elem

    def __repr__(self):
        return "QR-Invoice"

    def __str__(self):
        return "QR-Invoice"

    def debug(self):
        c = self.canv
        c.setStrokeColor("green")
        c.rect(5*mm, 5*mm, 52*mm, 95*mm)  # Empfangsschein
        c.rect(67*mm, 5*mm, 138*mm, 95*mm)  # Zahlteil
        c.rect(67*mm, 42*mm, 46*mm, 46*mm)  # QR-Code
        c.line(5*mm, 23*mm, 57*mm, 23*mm)
        c.line(5*mm, 37*mm, 57*mm, 37*mm)
        c.line(5*mm, 93*mm, 57*mm, 93*mm)
        c.line(67*mm, 93*mm, 118*mm, 93*mm)
        c.line(67*mm, 37*mm, 118*mm, 37*mm)
        c.line(67*mm, 15*mm, 205*mm, 15*mm)
        c.line(118*mm, 100*mm, 118*mm, 15*mm)
        c.setStrokeColor("black")

    def get_swiss_qr_payload(self):
        order = self.order
        recv = order.zahlungsempfaenger
        addr = order.rechnungsadresse
        qrpayload = []

        def ln(text=""):
            qrpayload.append(text)

        # QRCH
        # - Header
        # - - QRType
        ln("SPC")
        # - - Version
        ln("0200")
        # - - Coding
        ln("1")

        # - CdtrInf (Empfänger)
        # - - IBAN
        if recv.mode == "QRR":
            ln(recv.qriban.replace(" ", ""))
        else:
            ln(recv.iban.replace(" ", ""))
        # - - Cdtr
        # - - - AdrTp
        ln("K")
        # - - - Name
        ln(recv.firmenname)
        # - - - StrtNmOrAdrLine1
        ln(recv.adresszeile1)
        # - - - BldgNbOrAdrLine2
        ln(recv.adresszeile2)
        # - - - PstCd
        ln()
        # - - - TwnNm
        ln()
        # - - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln(recv.land)

        # - UltmtCdtr (Entgültiger Zahlungsempfänger)
        # - - AdrTp
        ln()
        # - - Name
        ln()
        # - - StrtNmOrAdrLine1
        ln()
        # - - BldgNbOrAdrLine2
        ln()
        # - - PstCd
        ln()
        # - - TwnNm
        ln()
        # - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln()

        # - CcyAmt
        # - - Amt
        ln(formatprice(order.fix_summe))
        # - - Ccy
        ln("CHF")

        # - UltmtDbtr (Entgültiger Zahlungspflichtiger)
        # - - AdrTp
        ln("K")
        # - - Name
        ln(addr["firma"] or f"{addr['vorname']} {addr['nachname']}")
        # - - StrtNmOrAdrLine1
        ln(addr["adresszeile1"])
        # - - BldgNbOrAdrLine2
        ln(addr["plz"]+" "+addr["ort"])
        # - - PstCd
        ln()
        # - - TwnNm
        ln()
        # - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln(addr["land"])

        # - RmtIn
        # - - TP
        ln(recv.mode)
        # - - Ref
        if recv.mode == "QRR":
            ln(order.referenznummer().replace(" ", ""))
        else:
            ln()
        # - - AddInf
        # - - - Ustrd
        ln(order.unstrukturierte_mitteilung())
        # - - - Trailer
        ln("EPD")
        # - - - StrdBkgInf
        ln(order.rechnungsinformationen())

        ## - AltPmtInf
        ## - - AltPmt
        # ln()
        ## - - AltPmt
        # ln()

        return "\n".join(qrpayload)

    def draw_qr_invoice(self):
        order = self.order
        invoiceinfo = order.rechnungsinformationen()
        strdbkginf = (invoiceinfo[:len(invoiceinfo)//2], invoiceinfo[len(invoiceinfo)//2:])
        ref = order.referenznummer()
        total = format(order.fix_summe, "08,.2f").replace(",", " ").lstrip(" 0")
        recv = order.zahlungsempfaenger
        addr = order.rechnungsadresse

        c = self.canv
        c.saveState()

        # QR-Code

        qrpayload = self.get_swiss_qr_payload()
        qr_code = qr.QrCodeWidget(qrpayload)
        qrbounds = qr_code.getBounds()
        qrwidth = qrbounds[2] - qrbounds[0]
        qrheight = qrbounds[3] - qrbounds[1]
        d = Drawing(
            52.2*mm, 52.2*mm, transform=[52.2*mm/qrwidth, 0, 0, 52.2*mm/qrheight, 0, 0])  # 46, 46
        d.add(qr_code)
        renderPDF.draw(d, c, 63.9*mm, 38.9*mm)  # 67, 42

        # Schweizerkreuz

        c.setFillColor("black")
        c.setStrokeColor("white")
        c.rect(86.5*mm, 61.5*mm, 7*mm, 7*mm, fill=1, stroke=1)
        c.setFillColor("white")
        c.rect(89.25*mm, 63*mm, 1.5*mm, 4*mm, fill=1, stroke=0)
        c.rect(88*mm, 64.25*mm, 4*mm, 1.5*mm, fill=1, stroke=0)

        c.setFillColor("black")
        c.setStrokeColor("black")

        # Begrenzungen Empfangsschein und Zahlteil und Abzutrennen-Hinweis

        if self.add_cut_info:
            c.line(0*mm, 105*mm, 210*mm, 105*mm)
            c.line(62*mm, 0*mm, 62*mm, 105*mm)
            c.setFont("Helvetica-Bold", 8)
            c.drawCentredString(
                105*mm, 107*mm, _("Vor der Einzahlung abzutrennen"))

        # Titel

        def titel(t, text, klein=False):
            t.setFont("Helvetica-Bold", 6 if klein else 8)
            t.textLine(text)
            t.moveCursor(0, 2)
            t.setFont("Helvetica", 8 if klein else 10)

        # Empfangsschein Angaben
        t = c.beginText(5*mm, 90*mm)
        titel(t, _("Konto / Zahlbar an"), True)
        t.textLine(recv.qriban if recv.mode == "QRR" else recv.iban)
        t.textLine(recv.firmenname)
        t.textLine(recv.adresszeile1)
        t.textLine(recv.adresszeile2)
        t.moveCursor(0, 9)
        if recv.mode == "QRR":
            titel(t, _("Referenz"), True)
            t.textLine(ref)
            t.moveCursor(0, 9)
        titel(t, _("Zahlbar durch"), True)
        t.textLine(addr["firma"] or f"{addr['vorname']} {addr['nachname']}")
        t.textLine(addr["adresszeile1"])
        t.textLine(f"{addr['plz']} {addr['ort']}")
        c.drawText(t)

        # Zahlteil Angaben
        t = c.beginText(118*mm, 97*mm)
        titel(t, _("Konto / Zahlbar an"))
        t.textLine(recv.qriban if recv.mode == "QRR" else recv.iban)
        t.textLine(recv.firmenname)
        t.textLine(recv.adresszeile1)
        t.textLine(recv.adresszeile2)
        t.moveCursor(0, 9)
        if recv.mode == "QRR":
            titel(t, _("Referenz"))
            t.textLine(ref)
            t.moveCursor(0, 9)
        titel(t, _("Zusätzliche Informationen"))
        t.textLine(order.unstrukturierte_mitteilung())
        t.textLine(strdbkginf[0])
        t.textLine(strdbkginf[1])
        t.moveCursor(0, 9)
        titel(t, _("Zahlbar durch"))
        t.textLine(addr["firma"] or f"{addr['vorname']} {addr['nachname']}")
        t.textLine(addr["adresszeile1"])
        t.textLine(f"{addr['plz']} {addr['ort']}")
        c.drawText(t)

        # Texte
        c.setFont("Helvetica-Bold", 11)
        c.drawString(5*mm, 97*mm, _("Empfangsschein"))
        c.drawString(67*mm, 97*mm, _("Zahlteil"))

        c.setFont("Helvetica-Bold", 6)
        c.drawString(5*mm, 33*mm, _("Währung"))
        c.drawString(20*mm, 33*mm, _("Betrag"))
        c.drawString(38*mm, 20*mm, _("Annahmestelle"))

        c.setFont("Helvetica", 8)
        c.drawString(5*mm, 30*mm, "CHF")
        c.drawString(20*mm, 30*mm, total)

        c.setFont("Helvetica-Bold", 8)
        c.drawString(67*mm, 33*mm, _("Währung"))
        c.drawString(87*mm, 33*mm, _("Betrag"))

        c.setFont("Helvetica", 10)
        c.drawString(67*mm, 29*mm, "CHF")
        c.drawString(87*mm, 29*mm, total)

        # c.setFont("Helvetica-Bold", 7)
        # c.drawString(67*mm, 11*mm, "Name AV1:")
        # c.drawString(67*mm, 8*mm, "Name AV2:")
        #
        # c.setFont("Helvetica", 7)
        # c.drawString(82*mm, 11*mm, "Linie 1")
        # c.drawString(82*mm, 8*mm, "Linie 2")

        # if settings.DEBUG:
        # self.debug()

        c.restoreState()

    def draw(self):
        self.canv.translate(-12*mm, -12*mm)
        self.draw_qr_invoice()


class _PDFOrderHeader(Flowable):
    @classmethod
    def from_order(cls, order, title, is_delivery_note=False):
        elem = cls()
        elem.width = 210
        elem.height = 75
        elem._fixedWidth = 210
        elem._fixedHeight = 75
        
        elem.order = order
        elem.title = str(title)
        elem.is_delivery_note = is_delivery_note
        return elem

    def draw_header(self):
        order = self.order
        ze = order.zahlungsempfaenger

        c = self.canv
        c.saveState()

        # Logo
        if ze.logourl:
            c.drawImage(ImageReader(ze.logourl), 120*mm, 67*mm,
                        width=20*mm, height=-20*mm, mask="auto", anchor="nw")

        # Company name
        c.setFont("Helvetica-Bold", 14)
        c.drawString(12*mm, 64*mm, ze.firmenname)

        # Company address
        t = c.beginText(12*mm, 57*mm)
        t.setFont("Helvetica", 10)
        t.textLine(ze.adresszeile1)
        if ze.land == "CH":
            t.textLine(ze.adresszeile2)
        else:
            t.textLine(f"{ze.land}-{ze.adresszeile2}")
        c.drawText(t)

        # Company info: Title
        t = c.beginText(12*mm, 46*mm)
        t.setFont("Helvetica", 8)
        t.textLine(_("Tel."))
        t.textLine(_("E-Mail"))
        if ze.webseite:
            t.textLine(_("Web"))
        if ze.firmenuid:
            t.textLine(_("MwSt"))
        c.drawText(t)

        # Company info: Content
        t = c.beginText(24*mm, 46*mm)
        t.setFont("Helvetica", 8)
        t.textLine(order.ansprechpartner.telefon)
        t.textLine(order.ansprechpartner.email)
        if ze.webseite:
            t.textLine(ze.webseite)
        if ze.firmenuid:
            t.textLine(ze.firmenuid)
        c.drawText(t)

        # Customer/Order info: Title
        t = c.beginText(12*mm, 27*mm)
        t.setFont("Helvetica", 12)
        t.textLine(_("Ihre Kundennummer"))
        t.textLine(_("Bestellnummer"))
        t.textLine(_("Bestelldatum"))
        if self.is_delivery_note:
            t.textLine(_("Datum"))
        else:
            t.textLine(_("Rechnungsdatum"))
        c.drawText(t)

        # Customer/Order info: Content
        t = c.beginText(64*mm, 27*mm)
        t.setFont("Helvetica", 12)
        t.textLine(order.kunde.pkfill(9) if order.kunde else "-"*9)
        t.textLine(order.pkfill(9))
        t.textLine(order.datum.strftime("%d.%m.%Y"))
        if self.is_delivery_note:
            t.textLine(datetime.now().date().strftime("%d.%m.%Y"))
        else:
            t.textLine(order.rechnungsdatum.strftime("%d.%m.%Y"))
        c.drawText(t)

        # Customer address
        t = c.beginText(120*mm, 27*mm)
        t.setFont("Helvetica", 12)
        if self.is_delivery_note:
            addr = order.lieferadresse
        else:
            addr = order.rechnungsadresse

        if addr["firma"]:
            t.textLine(addr["firma"])
        if addr["vorname"] or addr["nachname"]:
            t.textLine(f'{addr["vorname"]} {addr["nachname"]}'.strip())
        t.textLine(addr["adresszeile1"])
        if addr["adresszeile2"]:
            t.textLine(addr["adresszeile2"])
        t.textLine(f'{addr["plz"]} {addr["ort"]}'.strip())

        c.drawText(t)

        # Title and date
        c.setFont("Helvetica-Bold", 10)
        c.drawString(12*mm, 0*mm, self.title)

        c.setFont("Helvetica", 10)
        if len(self.title) <= 23:
            c.drawString(64*mm, 0*mm, f'{order.datum.year}-{order.pkfill(6)}' +
                         (f' (Online #{order.woocommerceid})' if order.woocommerceid else ''))
        else:
            c.drawString(120*mm, 0*mm, f'{order.datum.year}-{order.pkfill(6)}' +
                         (f' (Online #{order.woocommerceid})' if order.woocommerceid else ''))

        c.restoreState()

    def draw(self):
        self.canv.translate(-12*mm, -40*mm)
        self.draw_header()


class PDFOrder(PDFGenerator):
    def __init__(self, order, title, *, text=None, is_delivery_note=False, add_cut_lines=True, show_payment_conditions=None):
        order.fix_summe = order.summe_gesamt()

        lang = order.kunde.sprache if order.kunde and order.kunde.sprache else "de"

        # Header
        elements = [
            _PDFOrderHeader.from_order(
                order, title=title, is_delivery_note=is_delivery_note),
            Spacer(1, 48*mm),
        ]

        # Custom text
        if text:
            elements += [
                Paragraph(text.replace("\n", "\n<br />")),
                Spacer(1, 10*mm),
            ]

        # Main body
        if is_delivery_note:
            elements += [
                _PDFOrderProductTable.from_order(order, lang=lang)
            ]
        else:
            elements += [
                _PDFOrderPriceTable.from_order(order, lang=lang, show_payment_conditions=show_payment_conditions),
                Spacer(1, 65*mm),
                TopPadder(
                    _PDFOrderQrInvoice.from_order(
                        order, add_cut_lines=add_cut_lines
                    )
                ),
            ]

        # Set the elements
        self.elements = elements



# Generated by Django 4.1.5 on 2023-01-29 19:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("kmuhelper", "0086_rename_produktkategorie"),
    ]

    operations = [
        migrations.AddField(
            model_name="bestellung",
            name="bezahlt_am",
            field=models.DateField(
                blank=True, default=None, null=True, verbose_name="Bezahlt am"
            ),
        ),
        migrations.AddField(
            model_name="bestellung",
            name="versendet_am",
            field=models.DateField(
                blank=True, default=None, null=True, verbose_name="Versendet am"
            ),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.urls import reverse, reverse_lazy
from django.views.decorators.clickjacking import xframe_options_sameorigin as allow_iframe

from kmuhelper.utils import custom_app_list
from kmuhelper.modules.app.models import App_ToDo, App_Warenausgang, App_Zahlungseingang, App_Lagerbestand, App_Wareneingang

#####


@login_required(login_url=reverse_lazy("admin:login"))
def app_mobile_main(request):
    return render(request, "kmuhelper/app/mobile/main.html", {})


@allow_iframe
@login_required(login_url=reverse_lazy("admin:login"))
def app_mobile_home(request):
    return render(request, "kmuhelper/app/mobile/home.html", {
        "has_permission": True,
    })


@allow_iframe
def app_mobile_error(request):
    return render(request, "kmuhelper/app/mobile/error.html", {
        "has_permission": True,
    })


def app_mobile_manifest(request):
    response = render(request, "kmuhelper/app/mobile/manifest.webmanifest", {})
    response['Content-Type'] = 'text/json'
    response["Service-Worker-Allowed"] = "/"
    return response


@allow_iframe
@login_required(login_url=reverse_lazy("admin:login"))
def app_desktop(request):
    return render(request, 'admin/kmuhelper/_special/app/app_index.html', {
        'app_label': 'kmuhelper',
        'app_list': custom_app_list(request, [App_ToDo, App_Lagerbestand, App_Warenausgang, App_Wareneingang, App_Zahlungseingang], "KMUHelper App", reverse("kmuhelper:app-desktop")),
        'has_permission': True,
        'is_nav_sidebar_enabled': False,
        'is_popup': False,
        'title': 'KMUHelper App',
    })



from datetime import datetime, timedelta
from random import randint
from rich import print
import string


from django.db import models
from django.contrib import admin, messages
from django.core import mail
from django.core.validators import RegexValidator, MinValueValidator, MaxValueValidator
from django.utils import timezone
from django.utils.html import mark_safe, format_html
from django.urls import reverse

from kmuhelper import settings, constants
from kmuhelper.modules.emails.models import EMail, Attachment
from kmuhelper.modules.pdfgeneration import PDFOrder
from kmuhelper.overrides import CustomModel
from kmuhelper.utils import runden, formatprice, modulo10rekursiv, send_pdf, faq
from kmuhelper.translations import langselect, I18N_HELP_TEXT

from django.utils import translation
_ = translation.gettext

def log(string, *args):
    print("[deep_pink4][KMUHelper Main][/] -", string, *args)

###################


def defaultlieferungsname():
    return "Lieferung vom "+str(datetime.now().strftime("%d.%m.%Y"))


def defaultzahlungsempfaenger():
    if Zahlungsempfaenger.objects.exists():
        return Zahlungsempfaenger.objects.first().pk
    return None


def defaultansprechpartner():
    if Ansprechpartner.objects.exists():
        return Ansprechpartner.objects.first().pk
    return None


def defaultorderkey():
    return "kh-"+str(randint(10000000, 99999999))


def defaultzahlungskonditionen():
    return settings.get_db_setting("default-payment-conditions", "0:30")

# GUTSCHEINTYPEN = [
#     ("percent", "Prozent"),
#     ("fixed_cart", "Fixer Betrag auf den Warenkorb"),
#     ("fixed_product", "Fixer Betrag auf ein Produkt")
# ]

# ORDER_FREQUENCY_TYPES = [
#     ("weekly", "Wöchentlich"),
#     ("monthly", "Monatlich"),
#     ("yearly", "Jährlich"),
# ]

#############


class Ansprechpartner(CustomModel):
    """Model representing a contact person"""

    name = models.CharField(
        verbose_name="Name",
        max_length=50,
        help_text="Auf Rechnung ersichtlich!",
    )
    telefon = models.CharField(
        verbose_name="Telefon",
        max_length=50,
        help_text="Auf Rechnung ersichtlich!",
    )
    email = models.EmailField(
        verbose_name="E-Mail",
        help_text="Auf Rechnung ersichtlich!",
    )

    @admin.display(description="Ansprechpartner")
    def __str__(self):
        return f"{self.name} ({self.pk})"

    class Meta:
        verbose_name = "Ansprechpartner"
        verbose_name_plural = "Ansprechpartner"

    objects = models.Manager()

    admin_icon = "fas fa-user-tie"


class Bestellungskosten(CustomModel):
    """Model representing costs for an order (e.g. shipping costs)"""

    # Links to other models
    bestellung = models.ForeignKey(
        to='Bestellung',
        on_delete=models.CASCADE,
    )
    kosten = models.ForeignKey(
        to='Kosten',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        default=None,
    )

    # Custom data printed on the invoice
    bemerkung = models.CharField(
        verbose_name="Bemerkung",
        default="",
        max_length=250,
        blank=True,
        help_text="Wird auf die Rechnung gedruckt.",
    )

    rabatt = models.FloatField(
        verbose_name="Rabatt in %",
        default=0.0,
        validators=[
            MinValueValidator(0),
            MaxValueValidator(100),
        ],
    )

    # Data from Kosten
    name = models.CharField(
        verbose_name="Name",
        max_length=500,
        default="Zusätzliche Kosten",
        help_text=I18N_HELP_TEXT,
    )
    preis = models.FloatField(
        verbose_name="Preis (exkl. MwSt)",
        default=0.0,
    )
    mwstsatz = models.FloatField(
        verbose_name="MwSt-Satz",
        choices=constants.MWSTSETS,
        default=constants.MWST_DEFAULT,
    )

    # Calculated data
    def zwischensumme(self):
        return runden(self.preis*((100-self.rabatt)/100))

    def zwischensumme_ohne_rabatt(self):
        return runden(self.preis)

    def nur_rabatt(self):
        return runden(self.preis*(self.rabatt/100))*-1

    # Display methods
    @admin.display(description="Name", ordering="name")
    def clean_name(self):
        return langselect(self.name)

    @admin.display(description="Zwischensumme (exkl. MwSt)")
    def zwischensumme_display(self):
        return formatprice(self.zwischensumme()) + " CHF"

    @admin.display(description="Bestellungskosten")
    def __str__(self):
        if self.kosten:
            return f"({self.pk}) {self.clean_name()} ({self.kosten.pk})"
        return f"({self.pk}) {self.clean_name()}"

    def save(self, *args, **kwargs):
        if self.pk is None and self.kosten is not None:
            self.name = self.kosten.name
            self.preis = self.kosten.preis
            self.mwstsatz = self.kosten.mwstsatz
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Bestellungskosten"
        verbose_name_plural = "Bestellungskosten"

    objects = models.Manager()

    def copyto(self, bestellung):
        self.pk = None
        self._state.adding = True
        self.bestellung = bestellung
        self.save()


class Bestellungsposten(CustomModel):
    """Model representing the connection between 'Bestellung' and 'Produkt'"""

    bestellung = models.ForeignKey(
        to='Bestellung',
        on_delete=models.CASCADE,
    )
    produkt = models.ForeignKey(
        to='Produkt',
        on_delete=models.PROTECT,
    )
    bemerkung = models.CharField(
        verbose_name="Bemerkung",
        default="",
        max_length=250,
        blank=True,
        help_text="Wird auf die Rechnung gedruckt.",
    )

    menge = models.IntegerField(
        verbose_name="Menge",
        default=1,
    )
    rabatt = models.IntegerField(
        verbose_name="Rabatt in %",
        default=0,
        validators=[
            MinValueValidator(0),
            MaxValueValidator(100),
        ],
    )

    produktpreis = models.FloatField(
        verbose_name="Produktpreis (exkl. MwSt)",
        default=0.0,
    )

    # Calculated data
    def zwischensumme(self):
        return runden(self.produktpreis*self.menge*((100-self.rabatt)/100))

    def zwischensumme_ohne_rabatt(self):
        return runden(self.produktpreis*self.menge)

    def nur_rabatt(self):
        return runden(self.produktpreis*self.menge*(self.rabatt/100))*-1

    # Display methods
    @admin.display(description="MwSt-Satz")
    def mwstsatz(self):
        return formatprice(self.produkt.mwstsatz)

    @admin.display(description="Zwischensumme (exkl. MwSt)")
    def zwischensumme_display(self):
        return formatprice(self.zwischensumme()) + " CHF"

    @admin.display(description="Bestellungsposten")
    def __str__(self):
        return f'({self.pk}) {self.menge}x {self.produkt.clean_name()} ({self.produkt.pk})'

    def save(self, *args, **kwargs):
        if not self.produktpreis:
            self.produktpreis = runden(self.produkt.preis())
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Bestellungsposten"
        verbose_name_plural = "Bestellungsposten"

    objects = models.Manager()

    def copyto(self, bestellung):
        self.pk = None
        self._state.adding = True
        self.bestellung = bestellung
        self.save()


class Bestellung(CustomModel):
    """Model representing an order"""

    woocommerceid = models.IntegerField(
        verbose_name="WooCommerce ID",
        default=0,
    )

    datum = models.DateTimeField(
        verbose_name="Datum",
        default=timezone.now,
    )

    rechnungsdatum = models.DateField(
        verbose_name="Rechnungsdatum",
        default=None,
        blank=True,
        null=True,
        help_text="Datum der Rechnung. Wird auch als Startpunkt für die Zahlungskonditionen verwendet.",
    )
    rechnungstitel = models.CharField(
        verbose_name="Rechnungstitel",
        default="",
        blank=True,
        max_length=32,
        help_text="Titel der Rechnung. Leer lassen für 'RECHNUNG'",
    )
    rechnungstext = models.TextField(
        verbose_name="Rechnungstext",
        default="",
        blank=True,
        help_text=mark_safe(
            "Wird auf der Rechnung gedruckt! Unterstützt " +
            "<abbr title='<b>Fett</b><u>Unterstrichen</u><i>Kursiv</i>'>XML markup</abbr>."
        ),
    )
    zahlungskonditionen = models.CharField(
        verbose_name="Zahlungskonditionen",
        default=defaultzahlungskonditionen,
        validators=[
            RegexValidator(
                "^([0-9]+(\.[0-9]+)?:[0-9]+;)*0:[0-9]+$",
                "Bitte benutze folgendes Format: 'p:d;p:d' - p = Skonto in %; d = Tage",
            )
        ],
        max_length=16,
        help_text="Skonto und Zahlungsfrist -> "+faq('wie-funktionieren-zahlungskonditionen', 'Mehr erfahren'),
    )

    status = models.CharField(
        verbose_name="Status",
        max_length=11,
        default='processing',
        choices=constants.ORDERSTATUS,
    )
    versendet = models.BooleanField(
        verbose_name="Versendet",
        default=False,
        help_text='Mehr Infos -> '+faq('was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere', 'FAQ'),
    )
    versendet_am = models.DateField(
        verbose_name="Versendet am",
        default=None,
        blank=True,
        null=True,
    )
    trackingnummer = models.CharField(
        verbose_name="Trackingnummer",
        default="",
        blank=True,
        max_length=25,
        validators=[
            RegexValidator(
                r'^99\.[0-9]{2}\.[0-9]{6}\.[0-9]{8}$',
                'Bite benutze folgendes Format: 99.xx.xxxxxx.xxxxxxxx',
            )
        ],
        help_text="Bitte gib hier eine Trackingnummer der Schweizer Post ein. (optional)",
    )

    ausgelagert = models.BooleanField(
        verbose_name="Ausgelagert",
        default=False,
    )

    zahlungsmethode = models.CharField(
        verbose_name="Zahlungsmethode",
        max_length=7,
        default="cod",
        choices=constants.PAYMENTMETHODS,
    )
    bezahlt = models.BooleanField(
        verbose_name="Bezahlt",
        default=False,
        help_text='Mehr Infos -> '+faq('was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere', 'FAQ'),
    )
    bezahlt_am = models.DateField(
        verbose_name="Bezahlt am",
        default=None,
        blank=True,
        null=True,
    )

    kundennotiz = models.TextField(
        verbose_name="Kundennotiz",
        default="",
        blank=True,
        help_text="Vom Kunden erfasste Notiz.",
    )

    order_key = models.CharField(
        verbose_name="Bestellungs-Schlüssel",
        max_length=50,
        blank=True,
        default=defaultorderkey,
    )

    kunde = models.ForeignKey(
        to='Kunde',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='bestellungen',
    )
    zahlungsempfaenger = models.ForeignKey(
        to='Zahlungsempfaenger',
        on_delete=models.PROTECT,
        verbose_name="Zahlungsempfänger",
        default=defaultzahlungsempfaenger,
    )
    ansprechpartner = models.ForeignKey(
        to='Ansprechpartner',
        verbose_name="Ansprechpartner",
        on_delete=models.PROTECT,
        default=defaultansprechpartner,
    )

    # Rechnungsadresse

    @property
    def rechnungsadresse(self):
        return {
            'vorname': self.rechnungsadresse_vorname,
            'nachname': self.rechnungsadresse_nachname,
            'firma': self.rechnungsadresse_firma,
            'adresszeile1': self.rechnungsadresse_adresszeile1,
            'adresszeile2': self.rechnungsadresse_adresszeile2,
            'plz': self.rechnungsadresse_plz,
            'ort': self.rechnungsadresse_ort,
            'kanton': self.rechnungsadresse_kanton,
            'land': self.rechnungsadresse_land,
            'email': self.rechnungsadresse_email,
            'telefon': self.rechnungsadresse_telefon,
        }

    rechnungsadresse_vorname = models.CharField(
        verbose_name="Vorname",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_nachname = models.CharField(
        verbose_name="Nachname",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_firma = models.CharField(
        verbose_name="Firma",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_adresszeile1 = models.CharField(
        verbose_name="Adresszeile 1",
        max_length=250,
        default="",
        blank=True,
        help_text="Strasse und Hausnummer oder 'Postfach'",
    )
    rechnungsadresse_adresszeile2 = models.CharField(
        verbose_name="Adresszeile 2",
        max_length=250,
        default="",
        blank=True,
        help_text="Wird in QR-Rechnung NICHT verwendet!",
    )
    rechnungsadresse_ort = models.CharField(
        verbose_name="Ort",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_kanton = models.CharField(
        verbose_name="Kanton",
        max_length=50,
        default="",
        blank=True,
    )
    rechnungsadresse_plz = models.CharField(
        verbose_name="Postleitzahl",
        max_length=50,
        default="",
        blank=True,
    )
    rechnungsadresse_land = models.CharField(
        verbose_name="Land",
        max_length=2,
        default="CH",
        choices=constants.COUNTRIES,
    )
    rechnungsadresse_email = models.EmailField(
        verbose_name="E-Mail Adresse",
        blank=True,
    )
    rechnungsadresse_telefon = models.CharField(
        verbose_name="Telefon",
        max_length=50,
        default="",
        blank=True,
    )

    # Lieferadresse

    @property
    def lieferadresse(self):
        return {
            'vorname': self.lieferadresse_vorname,
            'nachname': self.lieferadresse_nachname,
            'firma': self.lieferadresse_firma,
            'adresszeile1': self.lieferadresse_adresszeile1,
            'adresszeile2': self.lieferadresse_adresszeile2,
            'plz': self.lieferadresse_plz,
            'ort': self.lieferadresse_ort,
            'kanton': self.lieferadresse_kanton,
            'land': self.lieferadresse_land,
            'email': self.lieferadresse_email,
            'telefon': self.lieferadresse_telefon,
        }

    lieferadresse_vorname = models.CharField(
        verbose_name="Vorname",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_nachname = models.CharField(
        verbose_name="Nachname",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_firma = models.CharField(
        verbose_name="Firma",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_adresszeile1 = models.CharField(
        verbose_name="Adresszeile 1",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_adresszeile2 = models.CharField(
        verbose_name="Adresszeile 2",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_ort = models.CharField(
        verbose_name="Ort",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_kanton = models.CharField(
        verbose_name="Kanton",
        max_length=50,
        default="",
        blank=True,
    )
    lieferadresse_plz = models.CharField(
        verbose_name="Postleitzahl",
        max_length=50,
        default="",
        blank=True,
    )
    lieferadresse_land = models.CharField(
        verbose_name="Land",
        max_length=2,
        default="CH",
        choices=constants.COUNTRIES,
    )
    lieferadresse_email = models.EmailField(
        verbose_name="E-Mail Adresse",
        blank=True,
    )
    lieferadresse_telefon = models.CharField(
        verbose_name="Telefon",
        max_length=50,
        default="",
        blank=True,
    )

    # Connections

    produkte = models.ManyToManyField(
        to='Produkt',
        through='Bestellungsposten',
        through_fields=('bestellung', 'produkt'),
    )

    kosten = models.ManyToManyField(
        to='Kosten',
        through='Bestellungskosten',
        through_fields=('bestellung', 'kosten'),
    )

    email_rechnung = models.ForeignKey(
        to='EMail',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='+',
    )
    email_lieferung = models.ForeignKey(
        to='EMail',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='+',
    )

    fix_summe = models.FloatField(
        verbose_name="Summe in CHF",
        default=0.0,
    )

    # # Wiederkehrende Rechnungen

    # recurring = models.BooleanField("Wiederkehrend", default=False)
    # recurring_next = models.DateField("Wiederkehrend am", default=None, blank=True, null=True)
    # recurring_until = models.DateField("Wiederkehrend bis", default=None, blank=True, null=True)
    # recurring_frequency = models.CharField("Häufigkeit", choices=ORDER_FREQUENCY_TYPES, default=None, blank=True, null=True)

    def import_customer_data(self):
        "Copy the customer data from the customer into the order"

        for field in constants.LIEFERADRESSE_FIELDS+constants.RECHNUNGSADRESSE_FIELDS:
            setattr(self, field, getattr(self.kunde, field))

    def second_save(self, *args, **kwargs):
        "This HAS to be called after all related models have been saved."

        self.fix_summe = self.summe_gesamt()
        if self.versendet and (not self.ausgelagert):
            for i in self.produkte.through.objects.filter(bestellung=self):
                i.produkt.lagerbestand -= i.menge
                i.produkt.save()
            self.ausgelagert = True

        super().save(*args, **kwargs)

    def save(self, *args, **kwargs):
        if not self.pk and not self.woocommerceid and self.kunde:
            self.import_customer_data()

        if self.rechnungsdatum is None:
            self.rechnungsdatum = timezone.now()

        super().save(*args, **kwargs)

    @admin.display(description="Trackinglink", ordering="trackingnummer")
    def trackinglink(self):
        return f'https://www.post.ch/swisspost-tracking?formattedParcelCodes={self.trackingnummer}' if self.trackingnummer else None

    def unstrukturierte_mitteilung(self):
        if self.zahlungsempfaenger.mode == "QRR":
            return str(self.datum.strftime("%d.%m.%Y"))
        return _("Referenznummer:")+" "+str(self.id)

    def referenznummer(self):
        a = self.pkfill(22)+"0000"
        b = a+str(modulo10rekursiv(a))
        c = b[0:2]+" "+b[2:7]+" "+b[7:12]+" " + \
            b[12:17]+" "+b[17:22]+" "+b[22:27]
        return c

    def rechnungsinformationen(self):
        date = (self.rechnungsdatum or self.datum).strftime("%y%m%d")

        output = f'//S1/10/{self.pk}/11/{date}'

        if self.zahlungsempfaenger.firmenuid:
            uid = self.zahlungsempfaenger.firmenuid.split("-")[1].replace(".", "")
            output += f'/30/{uid}'
        kond = self.zahlungskonditionen
        mwstdict = self.mwstdict()
        mwststring = ";".join(
            f'{satz}:{mwstdict[satz]}' for satz in mwstdict)

        output += f'/31/{date}/32/{mwststring}/40/{kond}'
        return output

    def get_paymentconditions(self):
        "Get the payment conditions as a list of dictionaries"

        data = []
        for pc in self.zahlungskonditionen.split(";"):
            percent, days = pc.split(":")
            percent, days = float(percent), int(days)
            data.append({
                "days": days,
                "date": self.rechnungsdatum+timedelta(days=days),
                "percent": percent,
                "price": runden(self.fix_summe*(1-(percent/100))),
            })
        data.sort(key=lambda x: x["date"])
        return data

    def mwstdict(self):
        "Get the VAT as a dictionary"
        mwst = {}
        for p in self.produkte.through.objects.filter(bestellung=self).select_related('produkt'):
            if str(p.produkt.mwstsatz) in mwst:
                mwst[str(p.produkt.mwstsatz)] += p.zwischensumme()
            else:
                mwst[str(p.produkt.mwstsatz)] = p.zwischensumme()
        for k in self.kosten.through.objects.filter(bestellung=self).select_related('kosten'):
            if str(k.mwstsatz) in mwst:
                mwst[str(k.mwstsatz)] += k.zwischensumme()
            else:
                mwst[str(k.mwstsatz)] = k.zwischensumme()
        for s in mwst:
            mwst[s] = runden(mwst[s])
        return mwst

    def summe(self):
        summe = 0
        for i in self.produkte.through.objects.filter(bestellung=self).select_related("produkt"):
            summe += i.zwischensumme()
        for i in self.kosten.through.objects.filter(bestellung=self).select_related("kosten"):
            summe += i.zwischensumme()
        return runden(summe)

    def summe_mwst(self):
        summe_mwst = 0
        mwstdict = self.mwstdict()
        for mwstsatz in mwstdict:
            summe_mwst += runden(float(mwstdict[mwstsatz]
                                       * (float(mwstsatz)/100)))
        return runden(summe_mwst)

    def summe_gesamt(self):
        return runden(self.summe()+self.summe_mwst())

    @admin.display(description="Rechnungstotal")
    def summeninfo(self):
        return f'{formatprice(self.summe())} CHF + {formatprice(self.summe_mwst())} CHF MwSt = {formatprice(self.summe_gesamt())} CHF'

    @admin.display(description="Total", ordering="fix_summe")
    def fix_summe_display(self):
        return f'{formatprice(self.fix_summe)} CHF'

    @admin.display(description="Name")
    def name(self):
        return (f'{self.datum.year}-' if self.datum and not isinstance(self.datum, str) else '') + \
               (f'{self.pkfill(6)}'+(f' (WC#{self.woocommerceid})' if self.woocommerceid else '')) + \
               (f' - {self.kunde}' if self.kunde is not None else " Gast")

    @admin.display(description="Info")
    def info(self):
        return f'{self.datum.strftime("%d.%m.%Y")} - '+((self.kunde.firma if self.kunde.firma else (f'{self.kunde.vorname} {self.kunde.nachname}')) if self.kunde else "Gast")

    @admin.display(description="Bezahlt nach")
    def bezahlt_nach_display(self):
        if self.bezahlt_am is None:
            return "-"
        
        daydiff = (self.bezahlt_am - self.rechnungsdatum).days
        return f'{daydiff} Tag{"en" if daydiff != 1 else ""}'

    @admin.display(description="Konditionen")
    def paymentconditions_display(self):
        "Get the payment conditions as a multiline string of values"

        conditions = self.get_paymentconditions()
        output = ""

        for condition in conditions:
            datestr = condition["date"].strftime("%d.%m.%Y")
            price = formatprice(condition["price"])
            percent = condition["percent"]
            output += f'{price} CHF bis {datestr} ({percent}%)<br>'

        return mark_safe(output)

    def is_correct_payment(self, amount: float, date: datetime):
        "Check if a payment made on a certain date has the correct amount for this order"

        if amount == self.fix_summe:
            return True
        
        for condition in self.get_paymentconditions():
            if amount == condition["price"] and date <= condition["date"]:
                return True

        return False

    @admin.display(description="Bestellung")
    def __str__(self):
        return self.name()

    def get_pdf(self, lieferschein: bool = False, digital: bool = True):
        return PDFOrder(self, is_delivery_note=lieferschein, add_cut_lines=digital).get_response(filename=('Lieferschein' if lieferschein else 'Rechnung')+' zu Bestellung '+str(self)+'.pdf')

    def create_email_rechnung(self):
        context = {
            "trackinglink": str(self.trackinglink()),
            "trackingdata": bool(self.trackingnummer and self.versendet),
            "id": str(self.id),
            "woocommerceid": str(self.woocommerceid),
            "woocommercedata": bool(self.woocommerceid),
        }

        self.email_rechnung = EMail.objects.create(
            subject=f"Ihre Rechnung Nr. { self.id }"+(
                f' (Online #{self.woocommerceid})' if self.woocommerceid else ''),
            to=self.rechnungsadresse_email,
            html_template="bestellung_rechnung.html",
            html_context=context,
            notes=f"Diese E-Mail wurde automatisch aus Bestellung #{self.pk} generiert.",
        )

        filename = f"Rechnung Nr. { self.id }"+(
            f' (Online #{self.woocommerceid})' if self.woocommerceid else '')+".pdf"

        self.email_rechnung.add_attachments(
            Attachment.objects.create_from_binary(
                filename=filename,
                content=PDFOrder(self, lieferschein=False,
                                 digital=True).get_pdf()
            )
        )

        self.save()
        return self.email_rechnung

    def create_email_lieferung(self):
        context = {
            "trackinglink": str(self.trackinglink()),
            "trackingdata": bool(self.trackinglink() and self.versendet),
            "id": str(self.id),
            "woocommerceid": str(self.woocommerceid),
            "woocommercedata": bool(self.woocommerceid),
        }

        self.email_lieferung = EMail.objects.create(
            subject=f"Ihre Lieferung Nr. { self.id }",
            to=self.lieferadresse_email,
            html_template="bestellung_lieferung.html",
            html_context=context,
            notes=f"Diese E-Mail wurde automatisch aus Bestellung #{self.pk} generiert.",
        )

        filename = f"Lieferschein Nr. { self.id }.pdf"

        self.email_lieferung.add_attachments(
            Attachment.objects.create_from_binary(
                filename=filename,
                content=PDFOrder(self, lieferschein=True,
                                 digital=True).get_pdf()
            )
        )

        self.save()
        return self.email_lieferung

    @admin.display(description="🔗 Notiz")
    def html_app_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_app_todo_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_app_todo_add") + \
                '?from_bestellung='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    @admin.display(description="🔗 Notiz")
    def html_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_notiz_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_notiz_add") + \
                '?from_bestellung='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    def get_stock_data(self):
        """Get the stock data of all products in this order"""

        return [p.get_stock_data() for p in self.produkte.all()]

    def email_stock_warning(self):
        email_receiver = settings.get_db_setting(
            "email-stock-warning-receiver")

        if email_receiver:
            warnings = []
            stock = self.get_stock_data()
            for data in stock:
                if data["stock_in_danger"]:
                    warnings.append(data)

            if warnings != []:
                email = EMail.objects.create(
                    subject="[KMUHelper] - Lagerbestand knapp!",
                    to=email_receiver,
                    html_template="bestellung_stock_warning.html",
                    html_context={
                        "warnings": warnings,
                    },
                    notes=f"Diese E-Mail wurde automatisch aus Bestellung #{self.pk} generiert.",
                )

                success = email.send(
                    headers={
                        "Bestellungs-ID": str(self.pk)
                    },
                )
                return bool(success)
        else:
            log("Keine E-Mail für Warnungen zum Lagerbestand festgelegt!")
        return None

    class Meta:
        verbose_name = "Bestellung"
        verbose_name_plural = "Bestellungen"

    def duplicate(self):
        new = Bestellung.objects.create(
            kunde=self.kunde,
            zahlungsempfaenger=self.zahlungsempfaenger,
            ansprechpartner=self.ansprechpartner,

            rechnungstitel=self.rechnungstitel,
            rechnungstext=self.rechnungstext,
            zahlungskonditionen=self.zahlungskonditionen,

            kundennotiz=f"Kopie aus Bestellung #{self.pk}\n--------------------------------\n{self.kundennotiz}",
        )

        for field in constants.LIEFERADRESSE_FIELDS+constants.RECHNUNGSADRESSE_FIELDS:
            setattr(new, field, getattr(self, field))

        for bp in self.produkte.through.objects.filter(bestellung=self):
            bp.copyto(new)
        for bk in self.kosten.through.objects.filter(bestellung=self):
            bk.copyto(new)
        new.save()
        return new

    def zu_lieferung(self):
        new = Lieferung.objects.create(
            name=f"Rücksendung von Bestellung #{self.pk}"
        )
        for lp in self.produkte.through.objects.filter(bestellung=self):
            new.produkte.add(lp.produkt, through_defaults={"menge": lp.menge})
        new.save()
        return new

    objects = models.Manager()

    admin_icon = "fas fa-clipboard-list"

    DICT_EXCLUDE_FIELDS = ['produkte', 'kosten', 'email_rechnung', 'email_lieferung', 'kunde',
                           'ansprechpartner', 'zahlungsempfaenger', 'ausgelagert',
                           'versendet', 'bezahlt', 'zahlungsmethode', 'order_key']


# class Gutschein(CustomModel):
#     woocommerceid = models.IntegerField('WooCommerce ID', default=0)
#
#     code = models.CharField("Gutscheincode", max_length=25)
#     menge = models.FloatField("Menge (Preis oder Anzahl Prozent)")
#     typ = models.CharField("Gutscheintyp", max_length=14, choices=GUTSCHEINTYPEN)
#     beschrieb = models.TextField("Beschrieb",default="",blank=True)
#     #datum_bis = models.DateField("Gültig bis", blank=True, null=True)
#     #nicht_kumulierbar = models.BooleanField("Nicht kumulierbar", default=True, help_text="Aktivieren, damit der Gutschein nicht kumuliert werden kann.")
#
#     produkte = models.ManyToManyField("Produkt", verbose_name="Produkte", help_text="Produkte, auf welche der Gutschein angewendet werden kann.")
#     ausgeschlossene_produkte = models.ManyToManyField("Produkt", verbose_name="Ausgeschlossenes Produkt", verbose_name_plural="Ausgeschlossene Produkte",help_text="Produkte, auf welche der Gutschein nicht angewendet werden kann.")
#
#     #limit_gesamt = models.IntegerField("Gesamtlimit", default=0, help_text="Anzahl Benutzungen")
#     #limit_pro_kunde = models.IntegerField("Limit pro Kunde", default=0, help_text="Anzahl Benutzungen pro Kunde")
#     limit_artikel = models.IntegerField("Limit an Artikeln", default=0, help_text="Anzahl Artikel, auf welche der Gutschein maximal angewendet werden kann.")
#
#     #kostenlose_lieferung = models.BooleanField("Kostenlose Lieferung", default=False, help_text="Wenn aktiviert, erlaube kostenlose Lieferung")
#
#     produktkategorien = models.ManyToManyField("Produktkategorie", verbose_name="Kategorie", verbose_name="Kategorien",help_text="Kategorien, auf welche der Gutschein angewendet werden kann.")
#     ausgeschlossene_produktkategorien = models.ManyToManyField("Kategorie", verbose_name="Ausgeschlossene Kategorie", verbose_name_plural="Ausgeschlossene Kategorie",help_text="Kategorien, auf welche der Gutschein nicht angewendet werden kann.")
#
#     nicht_kumulierbar_mit_aktion = models.BooleanField("Nicht kumulierbar mit Aktion", default=True, help_text="Aktivieren, damit der Gutschein nicht auf Produkte angewendet werden kann, welche in Aktion sind.")
#     mindestbetrag = models.FloatField("Mindestbetrag", default=0.0)
#     maximalbetrag = models.FloatField("Maximalbetrag", default=0.0)
#
#     #erlaubte_emails
#     #benutzt_von = models.ManyToManyField("Kunde")
#
#     def gueltig_fuer_bestellung(self, bestellung):
#         if ((self.mindestbetrag <= bestellung.summe_gesamt()) and (self.maximalbetrag == 0.0 or (self.maximalbetrag >= bestellung.summe_gesamt()))):
#             return True
#         else:
#             return False
#
#     def wert_fuer_bestellung(self, bestellung):
#         if self.gueltig_fuer_bestellung(bestellung):
#             if self.typ == "fixed_cart":
#                 return self.menge
#             elif self.typ == "fixed_product":
#                 pass
#             elif self.typ == "percent":
#                 pass
#         else:
#             return 0.0
#
#     class Meta:
#         verbose_name = "Gutschein"
#         verbose_name_plural = "Gutscheine"
#
#     objects = models.Manager()



class Kosten(CustomModel):
    """Model representing additional costs"""

    name = models.CharField(
        verbose_name="Name",
        max_length=500,
        default="Zusätzliche Kosten",
        help_text=I18N_HELP_TEXT,
    )
    preis = models.FloatField(
        verbose_name="Preis (exkl. MwSt)",
        default=0.0,
    )
    mwstsatz = models.FloatField(
        verbose_name="MwSt-Satz",
        choices=constants.MWSTSETS,
        default=constants.MWST_DEFAULT,
    )

    @admin.display(description="Name", ordering="name")
    def clean_name(self):
        return langselect(self.name)

    @admin.display(description="Kosten")
    def __str__(self):
        return f'{ self.clean_name() } ({ self.preis } CHF' + \
            (f' + {self.mwstsatz}% MwSt' if self.mwstsatz else '') + \
            f') ({self.pk})'

    class Meta:
        verbose_name = "Kosten"
        verbose_name_plural = "Kosten"

    objects = models.Manager()

    admin_icon = "fas fa-coins"


class Kunde(CustomModel):
    """Model representing a customer"""

    woocommerceid = models.IntegerField(
        verbose_name="WooCommerce ID",
        default=0,
    )

    email = models.EmailField(
        verbose_name="E-Mail Adresse",
        blank=True,
    )
    vorname = models.CharField(
        verbose_name="Vorname",
        max_length=250,
        default="",
        blank=True,
    )
    nachname = models.CharField(
        verbose_name="Nachname",
        max_length=250,
        default="",
        blank=True,
    )
    firma = models.CharField(
        verbose_name="Firma",
        max_length=250,
        default="",
        blank=True,
    )
    benutzername = models.CharField(
        verbose_name="Benutzername",
        max_length=50,
        default="",
        blank=True,
    )
    avatar_url = models.URLField(
        verbose_name="Avatar URL",
        blank=True,
        editable=False,
    )
    sprache = models.CharField(
        verbose_name="Sprache",
        default="de",
        choices=constants.LANGUAGES,
        max_length=2,
    )

    rechnungsadresse_vorname = models.CharField(
        verbose_name="Vorname",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_nachname = models.CharField(
        verbose_name="Nachname",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_firma = models.CharField(
        verbose_name="Firma",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_adresszeile1 = models.CharField(
        verbose_name="Adresszeile 1",
        max_length=250,
        default="",
        blank=True,
        help_text='Strasse und Hausnummer oder "Postfach"',
    )
    rechnungsadresse_adresszeile2 = models.CharField(
        verbose_name="Adresszeile 2",
        max_length=250,
        default="",
        blank=True,
        help_text="Wird in QR-Rechnung NICHT verwendet!",
    )
    rechnungsadresse_ort = models.CharField(
        verbose_name="Ort",
        max_length=250,
        default="",
        blank=True,
    )
    rechnungsadresse_kanton = models.CharField(
        verbose_name="Kanton",
        max_length=50,
        default="",
        blank=True,
    )
    rechnungsadresse_plz = models.CharField(
        verbose_name="Postleitzahl",
        max_length=50,
        default="",
        blank=True,
    )
    rechnungsadresse_land = models.CharField(
        verbose_name="Land",
        max_length=2,
        default="CH",
        choices=constants.COUNTRIES,
    )
    rechnungsadresse_email = models.EmailField(
        verbose_name="E-Mail Adresse",
        blank=True,
    )
    rechnungsadresse_telefon = models.CharField(
        verbose_name="Telefon",
        max_length=50,
        default="",
        blank=True,
    )

    lieferadresse_vorname = models.CharField(
        verbose_name="Vorname",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_nachname = models.CharField(
        verbose_name="Nachname",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_firma = models.CharField(
        verbose_name="Firma",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_adresszeile1 = models.CharField(
        verbose_name="Adresszeile 1",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_adresszeile2 = models.CharField(
        verbose_name="Adresszeile 2",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_ort = models.CharField(
        verbose_name="Ort",
        max_length=250,
        default="",
        blank=True,
    )
    lieferadresse_kanton = models.CharField(
        verbose_name="Kanton",
        max_length=50,
        default="",
        blank=True,
    )
    lieferadresse_plz = models.CharField(
        verbose_name="Postleitzahl",
        max_length=50,
        default="",
        blank=True,
    )
    lieferadresse_land = models.CharField(
        verbose_name="Land",
        max_length=2,
        default="CH",
        choices=constants.COUNTRIES,
    )
    lieferadresse_email = models.EmailField(
        verbose_name="E-Mail Adresse",
        blank=True,
    )
    lieferadresse_telefon = models.CharField(
        verbose_name="Telefon",
        max_length=50,
        default="",
        blank=True,
    )

    zusammenfuegen = models.ForeignKey(
        to='self',
        verbose_name="Zusammenfügen mit",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        help_text="Dies kann nicht widerrufen werden! Werte im aktuellen Kunden werden bevorzugt.",
    )
    webseite = models.URLField(
        verbose_name="Webseite",
        default="",
        blank=True,
    )
    bemerkung = models.TextField(
        verbose_name="Bemerkung",
        default="",
        blank=True,
    )

    email_registriert = models.ForeignKey(
        to='EMail',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    @admin.display(description="Avatar", ordering="avatar_url")
    def avatar(self):
        if self.avatar_url:
            return mark_safe('<img src="'+self.avatar_url+'" width="50px">')
        return ""

    @admin.display(description="Kunde")
    def __str__(self):
        s = f'{self.pkfill(8)} '
        if self.woocommerceid:
            s += f'(WC#{self.woocommerceid}) '
        if self.vorname:
            s += f'{self.vorname} '
        if self.nachname:
            s += f'{self.nachname} '
        if self.firma:
            s += f'{self.firma} '
        if self.rechnungsadresse_plz and self.rechnungsadresse_ort:
            s += f'({self.rechnungsadresse_plz} {self.rechnungsadresse_ort})'
        return s

    class Meta:
        verbose_name = "Kunde"
        verbose_name_plural = "Kunden"

    def save(self, *args, **kwargs):
        if self.zusammenfuegen:
            self.woocommerceid = self.woocommerceid or self.zusammenfuegen.woocommerceid

            self.email = self.email or self.zusammenfuegen.email
            self.vorname = self.vorname or self.zusammenfuegen.vorname
            self.nachname = self.nachname or self.zusammenfuegen.nachname
            self.firma = self.firma or self.zusammenfuegen.firma
            self.benutzername = self.benutzername or self.zusammenfuegen.benutzername
            self.avatar_url = self.avatar_url or self.zusammenfuegen.avatar_url
            self.sprache = self.sprache if self.sprache != "de" else self.zusammenfuegen.sprache

            self.rechnungsadresse_vorname = self.rechnungsadresse_vorname or self.zusammenfuegen.rechnungsadresse_vorname
            self.rechnungsadresse_nachname = self.rechnungsadresse_nachname or self.zusammenfuegen.rechnungsadresse_nachname
            self.rechnungsadresse_firma = self.rechnungsadresse_firma or self.zusammenfuegen.rechnungsadresse_firma
            self.rechnungsadresse_adresszeile1 = self.rechnungsadresse_adresszeile1 or self.zusammenfuegen.rechnungsadresse_adresszeile1
            self.rechnungsadresse_adresszeile2 = self.rechnungsadresse_adresszeile2 or self.zusammenfuegen.rechnungsadresse_adresszeile2
            self.rechnungsadresse_ort = self.rechnungsadresse_ort or self.zusammenfuegen.rechnungsadresse_ort
            self.rechnungsadresse_kanton = self.rechnungsadresse_kanton or self.zusammenfuegen.rechnungsadresse_kanton
            self.rechnungsadresse_plz = self.rechnungsadresse_plz or self.zusammenfuegen.rechnungsadresse_plz
            self.rechnungsadresse_land = self.rechnungsadresse_land or self.zusammenfuegen.rechnungsadresse_land
            self.rechnungsadresse_email = self.rechnungsadresse_email or self.zusammenfuegen.rechnungsadresse_email
            self.rechnungsadresse_telefon = self.rechnungsadresse_telefon or self.zusammenfuegen.rechnungsadresse_telefon

            self.lieferadresse_vorname = self.lieferadresse_vorname or self.zusammenfuegen.lieferadresse_vorname
            self.lieferadresse_nachname = self.lieferadresse_nachname or self.zusammenfuegen.lieferadresse_nachname
            self.lieferadresse_firma = self.lieferadresse_firma or self.zusammenfuegen.lieferadresse_firma
            self.lieferadresse_adresszeile1 = self.lieferadresse_adresszeile1 or self.zusammenfuegen.lieferadresse_adresszeile1
            self.lieferadresse_adresszeile2 = self.lieferadresse_adresszeile2 or self.zusammenfuegen.lieferadresse_adresszeile2
            self.lieferadresse_ort = self.lieferadresse_ort or self.zusammenfuegen.lieferadresse_ort
            self.lieferadresse_kanton = self.lieferadresse_kanton or self.zusammenfuegen.lieferadresse_kanton
            self.lieferadresse_plz = self.lieferadresse_plz or self.zusammenfuegen.lieferadresse_plz
            self.lieferadresse_land = self.lieferadresse_land or self.zusammenfuegen.lieferadresse_land
            self.lieferadresse_email = self.lieferadresse_email or self.zusammenfuegen.lieferadresse_email
            self.lieferadresse_telefon = self.lieferadresse_telefon or self.zusammenfuegen.lieferadresse_telefon

            self.webseite = self.webseite or self.zusammenfuegen.webseite
            self.bemerkung = self.bemerkung+"\n"+self.zusammenfuegen.bemerkung

            for bestellung in self.zusammenfuegen.bestellungen.all():
                bestellung.kunde = self
                bestellung.save()

            if getattr(self.zusammenfuegen, 'notiz', False):
                notiz = self.zusammenfuegen.notiz
                notiz.kunde = None if getattr(self, 'notiz', False) else self
                notiz.save()

            self.zusammenfuegen.delete()
            self.zusammenfuegen = None
        super().save(*args, **kwargs)

    def create_email_registriert(self):
        context = {
            "kunde": {
                "id": self.pk,
                "vorname": self.vorname,
                "nachname": self.nachname,
                "firma": self.firma,
                "email": self.email,
            }
        }

        self.email_registriert = EMail.objects.create(
            subject="Registrierung erfolgreich!",
            to=self.email,
            html_template="kunde_registriert.html",
            html_context=context,
            notes=f"Diese E-Mail wurde automatisch aus Kunde #{self.pk} generiert.",
        )

        self.save()
        return self.email_registriert

    @admin.display(description="🔗 Notiz")
    def html_app_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_app_todo_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_app_todo_add") + \
                '?from_kunde='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    @admin.display(description="🔗 Notiz")
    def html_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_notiz_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_notiz_add") + \
                '?from_kunde='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    objects = models.Manager()

    admin_icon = "fas fa-users"

    DICT_EXCLUDE_FIELDS = ['email_registriert', 'zusammenfuegen']


class Lieferant(CustomModel):
    """Model representing a supplier (used only for categorizing)"""

    kuerzel = models.CharField(
        verbose_name="Kürzel",
        max_length=5,
    )
    name = models.CharField(
        verbose_name="Name",
        max_length=50,
    )

    webseite = models.URLField(
        verbose_name="Webseite",
        blank=True,
    )
    telefon = models.CharField(
        verbose_name="Telefon",
        max_length=50,
        default="",
        blank=True,
    )
    email = models.EmailField(
        verbose_name="E-Mail",
        null=True,
        blank=True,
    )

    adresse = models.TextField(
        verbose_name="Adresse",
        default="",
        blank=True,
    )
    notiz = models.TextField(
        verbose_name="Notiz",
        default="",
        blank=True,
    )

    ansprechpartner = models.CharField(
        verbose_name="Ansprechpartner",
        max_length=250,
        default="",
        blank=True,
    )
    ansprechpartnertel = models.CharField(
        verbose_name="Ansprechpartner Telefon",
        max_length=50,
        default="",
        blank=True,
    )
    ansprechpartnermail = models.EmailField(
        verbose_name="Ansprechpartner E-Mail",
        null=True,
        default="",
        blank=True,
    )

    @admin.display(description="Lieferant")
    def __str__(self):
        return f'{self.name} ({self.pk})'

    def zuordnen(self):
        produkte = Produkt.objects.filter(lieferant=None)
        for produkt in produkte:
            produkt.lieferant = self
            produkt.save()
        return produkte.count()

    class Meta:
        verbose_name = "Lieferant"
        verbose_name_plural = "Lieferanten"

    objects = models.Manager()

    admin_icon = "fas fa-truck"


class Lieferungsposten(CustomModel):
    """Model representing the connection between 'Lieferung' and 'Produkt'"""

    lieferung = models.ForeignKey(
        to="Lieferung",
        on_delete=models.CASCADE,
    )
    produkt = models.ForeignKey(
        to="Produkt",
        on_delete=models.PROTECT,
    )
    menge = models.IntegerField(
        verbose_name="Menge",
        default=1,
    )

    @admin.display(description="Lieferungsposten")
    def __str__(self):
        return f'({self.lieferung.pk}) -> {self.menge}x {self.produkt}'

    class Meta:
        verbose_name = "Lieferungsposten"
        verbose_name_plural = "Lieferungsposten"

    objects = models.Manager()


class Lieferung(CustomModel):
    """Model representing an *incoming* delivery"""

    name = models.CharField(
        verbose_name="Name",
        max_length=50,
        default=defaultlieferungsname,
    )
    datum = models.DateField(
        verbose_name="Erfasst am",
        auto_now_add=True,
    )

    lieferant = models.ForeignKey(
        to="Lieferant",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    produkte = models.ManyToManyField(
        to="Produkt",
        through="Lieferungsposten",
        through_fields=("lieferung", "produkt"),
    )

    eingelagert = models.BooleanField(
        verbose_name="Eingelagert",
        default=False,
    )

    @admin.display(description="Anzahl Produklte")
    def anzahlprodukte(self):
        return self.produkte.through.objects.filter(lieferung=self).aggregate(models.Sum('menge'))["menge__sum"]

    def einlagern(self):
        if not self.eingelagert:
            for i in self.produkte.through.objects.filter(lieferung=self):
                i.produkt.lagerbestand += i.menge
                i.produkt.save()
            self.eingelagert = True
            self.save()
            return True
        return False

    @admin.display(description="🔗 Notiz")
    def html_app_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_app_todo_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_app_todo_add") + \
                '?from_lieferung='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    @admin.display(description="🔗 Notiz")
    def html_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_notiz_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_notiz_add") + \
                '?from_lieferung='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    @admin.display(description="Lieferung")
    def __str__(self):
        return f'{self.name} ({self.pk})'

    class Meta:
        verbose_name = "Lieferung"
        verbose_name_plural = "Lieferungen"

    objects = models.Manager()

    admin_icon = "fas fa-truck-ramp-box"


class Notiz(CustomModel):
    """Model representing a note"""

    name = models.CharField(
        verbose_name="Name",
        max_length=50,
    )
    beschrieb = models.TextField(
        verbose_name="Beschrieb",
        default="",
        blank=True,
    )

    erledigt = models.BooleanField(
        verbose_name="Erledigt",
        default=False,
    )

    priority = models.IntegerField(
        verbose_name="Priorität",
        default=0,
        blank=True,
    )
    erstellt_am = models.DateTimeField(
        verbose_name="Erstellt am",
        auto_now_add=True,
    )

    bestellung = models.OneToOneField(
        to="Bestellung",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="notiz",
    )
    produkt = models.OneToOneField(
        to="Produkt",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="notiz",
    )
    kunde = models.OneToOneField(
        to="Kunde",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="notiz",
    )
    lieferung = models.OneToOneField(
        to="Lieferung",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="notiz",
    )

    @admin.display(description="🔗 Notiz")
    def __str__(self):
        return f'{self.name} ({self.pk})'

    def links(self):
        text = ""
        if self.bestellung:
            url = reverse("admin:kmuhelper_bestellung_change",
                          kwargs={"object_id": self.bestellung.pk})
            text += f"Bestellung <a href='{url}'>#{self.bestellung.pk}</a><br>"
        if self.produkt:
            url = reverse("admin:kmuhelper_produkt_change",
                          kwargs={"object_id": self.produkt.pk})
            text += f"Produkt <a href='{url}'>#{self.produkt.pk}</a><br>"
        if self.kunde:
            url = reverse("admin:kmuhelper_kunde_change",
                          kwargs={"object_id": self.kunde.pk})
            text += f"Kunde <a href='{url}'>#{self.kunde.pk}</a><br>"
        if self.lieferung:
            url = reverse("admin:kmuhelper_lieferung_change",
                          kwargs={"object_id": self.lieferung.pk})
            text += f"Lieferung <a href='{url}'>#{self.lieferung.pk}</a><br>"
        return mark_safe(text) or "Diese Notiz hat keine Verknüpfungen."

    class Meta:
        verbose_name = "Notiz"
        verbose_name_plural = "Notizen"

    objects = models.Manager()

    admin_icon = "fas fa-note-sticky"


class Produkt(CustomModel):
    """Model representing a product"""

    artikelnummer = models.CharField(
        verbose_name="Artikelnummer",
        max_length=25,
    )

    woocommerceid = models.IntegerField(
        verbose_name='WooCommerce ID',
        default=0,
    )

    name = models.CharField(
        verbose_name='Name',
        max_length=500,
        help_text=I18N_HELP_TEXT,
    )
    kurzbeschrieb = models.TextField(
        verbose_name='Kurzbeschrieb',
        default="",
        blank=True,
        help_text=I18N_HELP_TEXT,
    )
    beschrieb = models.TextField(
        verbose_name='Beschrieb',
        default="",
        blank=True,
        help_text=I18N_HELP_TEXT,
    )

    mengenbezeichnung = models.CharField(
        verbose_name='Mengenbezeichnung',
        max_length=100,
        default="Stück",
        blank=True,
        help_text=I18N_HELP_TEXT,
    )
    verkaufspreis = models.FloatField(
        verbose_name='Normalpreis in CHF (exkl. MwSt)',
        default=0,
    )
    mwstsatz = models.FloatField(
        verbose_name='Mehrwertsteuersatz',
        choices=constants.MWSTSETS,
        default=constants.MWST_DEFAULT,
    )

    lagerbestand = models.IntegerField(
        verbose_name="Lagerbestand",
        default=0,
    )
    soll_lagerbestand = models.IntegerField(
        verbose_name="Soll-Lagerbestand",
        default=1,
    )

    bemerkung = models.TextField(
        verbose_name='Bemerkung',
        default="",
        blank=True,
        help_text="Wird nicht gedruckt oder angezeigt; nur für eigene Zwecke.",
    )

    aktion_von = models.DateTimeField(
        verbose_name="In Aktion von",
        blank=True,
        null=True,
    )
    aktion_bis = models.DateTimeField(
        verbose_name="In Aktion bis",
        blank=True,
        null=True,
    )
    aktion_preis = models.FloatField(
        verbose_name="Aktionspreis in CHF (exkl. MwSt)",
        blank=True,
        null=True,
    )

    datenblattlink = models.CharField(
        verbose_name='Datenblattlink',
        max_length=500,
        default="",
        blank=True,
    )
    bildlink = models.URLField(
        verbose_name='Bildlink',
        default="",
        blank=True,
    )

    lieferant = models.ForeignKey(
        to="Lieferant",
        verbose_name="Lieferant",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    lieferant_preis = models.CharField(
        verbose_name="Lieferantenpreis",
        default="",
        blank=True,
        max_length=20,
    )
    lieferant_artikelnummer = models.CharField(
        verbose_name="Lieferantenartikelnummer",
        default="",
        blank=True,
        max_length=25,
    )
    lieferant_url = models.URLField(
        verbose_name="Lieferantenurl (Für Nachbestellungen)",
        default="",
        blank=True,
    )

    kategorien = models.ManyToManyField(
        to="Produktkategorie",
        through="ProduktProduktkategorieVerbindung",
        through_fields=("produkt", "kategorie"),
        verbose_name="Kategorie",
        related_name="produkte",
    )

    @admin.display(description="Name", ordering="name")
    def clean_name(self, lang="de"):
        return langselect(self.name, lang)

    @admin.display(description="Kurzbeschrieb", ordering="kurzbeschrieb")
    def clean_kurzbeschrieb(self, lang="de"):
        return langselect(self.kurzbeschrieb, lang)

    @admin.display(description="Beschrieb", ordering="beschrieb")
    def clean_beschrieb(self, lang="de"):
        return langselect(self.beschrieb, lang)

    @admin.display(description="In Aktion", boolean=True)
    def in_aktion(self, zeitpunkt: datetime = None):
        zp = zeitpunkt or timezone.now()
        if self.aktion_von and self.aktion_bis and self.aktion_preis:
            return bool((self.aktion_von < zp) and (zp < self.aktion_bis))
        return False

    @admin.display(description="Aktueller Preis in CHF (exkl. MwSt)")
    def preis(self, zeitpunkt: datetime = None):
        zp = zeitpunkt or timezone.now()
        return self.aktion_preis if self.in_aktion(zp) else self.verkaufspreis

    @admin.display(description="Bild", ordering="bildlink")
    def bild(self):
        if self.bildlink:
            return format_html('<img src="{}" width="100px">', self.bildlink)
        return ""

    def get_reserved_stock(self):
        n = 0
        for bp in Bestellungsposten.objects.filter(bestellung__versendet=False, produkt__id=self.id):
            n += bp.menge
        return n

    def get_incoming_stock(self):
        n = 0
        for lp in Lieferungsposten.objects.filter(lieferung__eingelagert=False, produkt__id=self.id):
            n += lp.menge
        return n

    def get_stock_data(self, includemessage=False):
        """Get the stock and product information as a dictionary"""

        p_id = self.id
        p_name = self.clean_name()
        p_artikelnummer = self.artikelnummer

        n_current = self.lagerbestand
        n_going = self.get_reserved_stock()
        n_coming = self.get_incoming_stock()
        n_min = self.soll_lagerbestand

        data = {
            "product": {
                "id": p_id,
                "artikelnummer": p_artikelnummer,
                "name": p_name,
            },
            "stock": {
                "current": n_current,
                "going": n_going,
                "coming": n_coming,
                "min": n_min,
            },
            "stock_overbooked": n_current-n_going < 0,
            "stock_in_danger": n_current-n_going < n_min,
        }

        if includemessage:
            stockstring = f"Aktuell: { n_current } | Ausgehend: { n_going }" + (
                f" | Eingehend: { n_coming }" if n_coming else "")
            adminurl = reverse(
                f'admin:kmuhelper_produkt_change', args=[self.pk])
            adminlink = format_html('<a href="{}">{}</a>', adminurl, p_name)

            formatdata = (adminlink, p_artikelnummer, stockstring)

            if data["stock_overbooked"]:
                data["message"] = format_html(
                    'Zu wenig Lagerbestand bei "{}" [{}]: {}', *formatdata)
            elif data["stock_in_danger"]:
                data["message"] = format_html(
                    'Knapper Lagerbestand bei "{}" [{}]: {}', *formatdata)

        return data

    def show_stock_warning(self, request):
        data = self.get_stock_data(includemessage=True)

        if data["stock_overbooked"]:
            messages.error(request, data["message"])
        elif data["stock_in_danger"]:
            messages.warning(request, data["message"])

    @admin.display(description="🔗 Notiz")
    def html_app_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_app_todo_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_app_todo_add") + \
                '?from_produkt='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    @admin.display(description="🔗 Notiz")
    def html_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_notiz_change",
                           kwargs={"object_id": self.notiz.pk})
            text = "Notiz ansehen"
        else:
            link = reverse("admin:kmuhelper_notiz_add") + \
                '?from_produkt='+str(self.pk)
            text = "Notiz hinzufügen"
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    @admin.display(description="Produkt")
    def __str__(self):
        return f'{self.artikelnummer} - {self.clean_name()} ({self.pk})'

    class Meta:
        verbose_name = "Produkt"
        verbose_name_plural = "Produkte"

    objects = models.Manager()

    admin_icon = "fas fa-cubes"


class Produktkategorie(CustomModel):
    """Model representing a category for products"""

    woocommerceid = models.IntegerField(
        verbose_name="WooCommerce ID",
        default=0,
    )

    name = models.CharField(
        verbose_name="Name",
        max_length=250,
        default="",
    )
    beschrieb = models.TextField(
        verbose_name="Beschrieb",
        default="",
        blank=True,
    )
    bildlink = models.URLField(
        verbose_name="Bildlink",
        blank=True,
    )

    uebergeordnete_kategorie = models.ForeignKey(
        to='self',
        verbose_name="Übergeordnete Kategorie",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    @admin.display(description="Bild", ordering="bildlink")
    def bild(self):
        if self.bildlink:
            return format_html('<img src="{}" width="100px">', self.bildlink)
        return ""

    @admin.display(description="Anzahl Produkte")
    def anzahl_produkte(self):
        return self.produkte.count()

    @admin.display(description="Name", ordering="name")
    def clean_name(self):
        return langselect(self.name)

    @admin.display(description="Beschrieb", ordering="beschrieb")
    def clean_beschrieb(self):
        return langselect(self.beschrieb)

    @admin.display(description="Kategorie")
    def __str__(self):
        return f'{self.clean_name()} ({self.pk})'

    class Meta:
        verbose_name = "Produktkategorie"
        verbose_name_plural = "Produktkategorien"

    objects = models.Manager()

    admin_icon = "fas fa-folder-tree"

class ProduktProduktkategorieVerbindung(CustomModel):
    """Model representing the connection between 'Produkt' and 'Produktkategorie'"""

    produkt = models.ForeignKey("Produkt", on_delete=models.CASCADE)
    kategorie = models.ForeignKey("Produktkategorie", on_delete=models.CASCADE)

    @admin.display(description="Produktkategorie")
    def __str__(self):
        return f'({self.kategorie.pk}) {self.kategorie.clean_name()} <-> {self.produkt}'

    class Meta:
        verbose_name = "Produkt-Kategorie-Verknüpfung"
        verbose_name_plural = "Produkt-Kategorie-Verknüpfungen"

    objects = models.Manager()


class Zahlungsempfaenger(CustomModel):
    """Model representing a payment receiver for the qr bill"""

    mode = models.CharField(
        verbose_name="Modus",
        max_length=15,
        choices=[
            ('QRR', "QR-Referenz"),
            ('NON', "Ohne Referenz"),
        ],
        default='QRR'
    )

    qriban = models.CharField(
        verbose_name="QR-IBAN",
        max_length=21+5,
        validators=[
            RegexValidator(
                r'^(CH|LI)[0-9]{2}\s3[0-9]{3}\s[0-9]{4}\s[0-9]{4}\s[0-9]{4}\s[0-9]{1}$',
                'Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x',
            ),
        ],
        help_text="QR-IBAN mit Leerzeichen (Nur verwendet im Modus 'QR-Referenz')",
        blank=True,
        default="",
    )
    iban = models.CharField(
        verbose_name="IBAN",
        max_length=21+5,
        validators=[
            RegexValidator(
                r'^(CH|LI)[0-9]{2}\s[0-9]{4}\s[0-9]{4}\s[0-9]{4}\s[0-9]{4}\s[0-9]{1}$',
                'Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x',
            ),
        ],
        help_text="IBAN mit Leerzeichen (Nur verwendet im Modus 'Ohne Referenz')",
        blank=True,
        default="",
    )

    logourl = models.URLField(
        verbose_name="Logo (URL)",
        validators=[
            RegexValidator(
                r'''^[0-9a-zA-Z\-\.\|\?\(\)\*\+&"'_:;/]+\.(png|jpg)$''',
                '''Nur folgende Zeichen gestattet: 0-9a-zA-Z-_.:;/|?&()"'*+ - ''' +
                '''Muss auf .jpg/.png enden.''',
            ),
        ],
        help_text="URL eines Bildes (.jpg/.png) - Wird auf die Rechnung gedruckt.",
        blank=True,
        default="",
    )
    firmenname = models.CharField(
        verbose_name="Firmennname",
        max_length=70,
        help_text="Name der Firma oder des Empfängers",
    )
    firmenuid = models.CharField(
        verbose_name="Firmen-UID",
        max_length=15,
        validators=[
            RegexValidator(
                r'^CHE-[0-9]{3}\.[0-9]{3}\.[0-9]{3}$',
                'Bite benutze folgendes Format: CHE-123.456.789'
            )
        ],
        help_text="UID der Firma - Format: CHE-123.456.789 (Mehrwertsteuernummer)",
        blank=True,
        default="",
    )
    adresszeile1 = models.CharField(
        verbose_name="Strasse und Hausnummer oder 'Postfach'",
        max_length=70,
    )
    adresszeile2 = models.CharField(
        verbose_name="PLZ und Ort",
        max_length=70,
    )
    land = models.CharField(
        verbose_name="Land",
        max_length=2,
        choices=constants.COUNTRIES,
        default="CH",
    )
    email = models.EmailField(
        verbose_name="E-Mail",
        default="",
        blank=True,
        help_text="Nicht auf der Rechnung ersichtlich",
    )
    telefon = models.CharField(
        verbose_name="Telefon",
        max_length=70,
        default="",
        blank=True,
        help_text="Nicht auf der Rechnung ersichtlich",
    )
    webseite = models.URLField(
        verbose_name="Webseite",
        default="",
        blank=True,
        help_text="Auf der Rechnung ersichtlich, sofern vorhanden!",
    )

    @classmethod
    def _check_iban(cls, iban: str):
        try:
            b = ''
            for i in (0, 1):
                a = str(iban)[i].upper()
                if a not in string.ascii_uppercase:
                    return False
                b += str(ord(a)-55)
            Nr = ''.join([z for z in str(iban)
                          [2:] if z in string.digits])
            return int(int(Nr[2:] + b + Nr[:2]) % 97) == 1
        except IndexError:
            return False

    def has_valid_qr_iban(self):
        return self._check_iban(self.qriban)

    def has_valid_iban(self):
        return self._check_iban(self.iban)

    def has_valid_uid(self):
        try:
            u = self.firmenuid.split("-")[1].replace(".", "")
            p = 11 - (((int(u[0])*5)+(int(u[1])*4)+(int(u[2])*3)+(int(u[3])*2) +
                       (int(u[4])*7)+(int(u[5])*6)+(int(u[6])*5)+(int(u[7])*4)) % 11)
            return int(u[8]) == p
        except Exception as e:
            log("Error while validating UID:", e)
            return False

    @admin.display(description="Zahlungsempfänger")
    def __str__(self):
        return f'{self.firmenname} ({self.pk})'

    class Meta:
        verbose_name = "Zahlungsempfänger"
        verbose_name_plural = "Zahlungsempfänger"

    objects = models.Manager()

    admin_icon = "fas fa-hand-holding-dollar"



# Generated by Django 3.0.4 on 2020-07-20 17:36

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0029_notiz_produkt'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='kunde',
            name='notiz',
        ),
        migrations.AddField(
            model_name='kunde',
            name='bemerkung',
            field=models.TextField(blank=True, default='', verbose_name='Bemerkung'),
        ),
        migrations.AddField(
            model_name='notiz',
            name='kunde',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notiz', to='kmuhelper.Kunde'),
        ),
        migrations.AlterField(
            model_name='produkt',
            name='mengenbezeichnung',
            field=models.CharField(blank=True, default='[:de]Stück[:fr]Pièce[:it]Pezzo[:en]Piece[:]', max_length=100, verbose_name='Mengenbezeichnung'),
        ),
    ]



# Generated by Django 3.0.4 on 2020-05-31 14:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0010_kunde_webseite'),
    ]

    operations = [
        migrations.AddField(
            model_name='kunde',
            name='firma',
            field=models.CharField(blank=True, default='', max_length=50, verbose_name='Firma'),
        ),
    ]



# Generated by Django 4.1.5 on 2023-02-03 18:12

import django.core.validators
from django.db import migrations, models
import kmuhelper.modules.main.models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0087_bestellung_bezahlt_am_bestellung_versendet_am'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bestellung',
            name='bezahlt',
            field=models.BooleanField(default=False, help_text='Mehr Infos -> <a href="https://rafaelurben.github.io/django-kmuhelper/manual/faq#was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere" target="_blank">FAQ</a>', verbose_name='Bezahlt'),
        ),
        migrations.AlterField(
            model_name='bestellung',
            name='versendet',
            field=models.BooleanField(default=False, help_text='Mehr Infos -> <a href="https://rafaelurben.github.io/django-kmuhelper/manual/faq#was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere" target="_blank">FAQ</a>', verbose_name='Versendet'),
        ),
        migrations.AlterField(
            model_name='bestellung',
            name='zahlungskonditionen',
            field=models.CharField(default=kmuhelper.modules.main.models.defaultzahlungskonditionen, help_text='Skonto und Zahlungsfrist -> <a href="https://rafaelurben.github.io/django-kmuhelper/manual/faq#wie-funktionieren-zahlungskonditionen" target="_blank">Mehr erfahren</a>', max_length=16, validators=[django.core.validators.RegexValidator('^([0-9]+(\\.[0-9]+)?:[0-9]+;)*0:[0-9]+$', "Bitte benutze folgendes Format: 'p:d;p:d' - p = Skonto in %; d = Tage")], verbose_name='Zahlungskonditionen'),
        ),
    ]



import csv

from django.core.management.base import BaseCommand, CommandError

from kmuhelper.modules.main.models import Kunde
from kmuhelper.utils import getfirstindex


class Command(BaseCommand):
    help = 'Imports customers from a .csv file.'

    def add_arguments(self, parser):
        parser.add_argument(
            'filepath', help="Absolute filepath to the .csv file", type=str)

    def handle(self, *args, **options):
        self.stdout.write(self.style.SUCCESS("Import gestartet!"))
        try:
            with open(options["filepath"], newline='', encoding="utf-8-sig") as f:
                reader = csv.reader(f)
                rows = list(reader)

            indexrow = rows.pop(0)

            ###
            index_id = getfirstindex(indexrow, ["Nummer", "ID"])

            index_email = getfirstindex(
                indexrow, ["Email", "E-Mail", "EMail", "EMAIL"])
            index_vorname = getfirstindex(
                indexrow, ["Vorname"])
            index_nachname = getfirstindex(
                indexrow, ["Name", "Nachname"])
            index_sprache = getfirstindex(
                indexrow, ["Sprache", "SPRACHE"])

            index_firma = getfirstindex(
                indexrow, ["Firma"])
            index_strasse = getfirstindex(
                indexrow, ["Strasse"])
            index_ort = getfirstindex(
                indexrow, ["Ort"])
            index_plz = getfirstindex(
                indexrow, ["PLZ", "Plz", "Postleitzahl"])
            index_land = getfirstindex(
                indexrow, ["Land"])
            index_telefon = getfirstindex(
                indexrow, ["Telefon", "Telefon1"])

            index_adresszeile2 = getfirstindex(
                indexrow, ["Adresszeile2", "Zusatz"])

            index_notiz = getfirstindex(
                indexrow, ["Notiz"])
            index_web = getfirstindex(
                indexrow, ["Webseite", "Homepage"])

            ###

            for row in rows:
                objid = row[index_id] if index_id is not None else None

                email = row[index_email] if index_email is not None else ""
                vorname = row[index_vorname] if index_vorname is not None else ""
                nachname = row[index_nachname] if index_nachname is not None else ""
                sprache = ("de" if row[index_sprache].upper() in ["D", "DE", "DEUTSCH"] else
                           ("fr" if row[index_sprache].upper() in ["F", "FR", "FRANZÖSISCH"] else
                            ("it" if row[index_sprache].upper() in ["I", "IT", "ITALIENISCH"] else
                             ("en" if row[index_sprache].upper() in ["E", "EN", "ENGLISCH"] else
                              "de")))) if index_sprache is not None else "de"

                firma = row[index_firma] if index_firma is not None else ""
                adresszeile1 = row[index_strasse] if index_strasse is not None else ""
                adresszeile2 = row[index_adresszeile2] if index_adresszeile2 is not None else ""
                ort = row[index_ort] if index_ort is not None else ""
                plz = row[index_plz] if index_plz is not None else ""
                land = row[index_land] if index_land is not None else ""
                telefon = row[index_telefon] if index_telefon is not None else ""

                notiz = row[index_notiz] if index_notiz is not None else ""
                webseite = row[index_web] if index_web is not None else ""

                oldwithid = Kunde.objects.filter(pk=objid).exists()
                oldwithemail = Kunde.objects.filter(
                    email=email).exists() if email else False
                oldwithidandemail = Kunde.objects.filter(
                    pk=objid, email=email).exists()

                self.stdout.write(self.style.SUCCESS(""))

                if not oldwithid and not oldwithemail:
                    self.stdout.write(self.style.SUCCESS("Kunden erstellen:" +
                                                         "\n  ID:         " + str(objid) +
                                                         "\n  E-Mail:     " + email +
                                                         "\n  Vorname:    " + vorname +
                                                         "\n  Nachname:   " + nachname +
                                                         "\n  Firma:      " + firma
                                                         ))
                    Kunde.objects.create(
                        pk=objid,

                        email=email,
                        vorname=vorname,
                        nachname=nachname,
                        firma=firma,
                        sprache=sprache,

                        rechnungsadresse_vorname=vorname,
                        rechnungsadresse_nachname=nachname,
                        rechnungsadresse_firma=firma,
                        rechnungsadresse_adresszeile1=adresszeile1,
                        rechnungsadresse_adresszeile2=adresszeile2,
                        rechnungsadresse_ort=ort,
                        #rechnungsadresse_kanton = "",
                        rechnungsadresse_plz=plz,
                        rechnungsadresse_land=land,
                        rechnungsadresse_email=email,
                        rechnungsadresse_telefon=telefon,

                        lieferadresse_vorname=vorname,
                        lieferadresse_nachname=nachname,
                        lieferadresse_firma=firma,
                        lieferadresse_adresszeile1=adresszeile1,
                        lieferadresse_adresszeile2=adresszeile2,
                        lieferadresse_ort=ort,
                        #lieferadresse_kanton = "",
                        lieferadresse_plz=plz,
                        lieferadresse_land=land,

                        webseite=webseite,
                        notiz=notiz
                    )
                elif oldwithidandemail:
                    self.stdout.write(self.style.SUCCESS("Kunde existiert bereits!" +
                                                         "\n  ID:         " + str(objid) +
                                                         "\n  E-Mail:     " + email +
                                                         "\n  Vorname:    " + vorname +
                                                         "\n  Nachname:   " + nachname +
                                                         "\n  Firma:      " + firma
                                                         ))
                elif oldwithemail and not oldwithid:
                    old = Kunde.objects.get(email=email)
                    self.stdout.write(self.style.SUCCESS("Kunde mit der E-Mail '"+email+"' existiert bereits, jedoch ist die ID unterschiedlich!" +
                                                         "\n  KMUHelper-Daten:" +
                                                         "\n    ID:       " + str(old.pk) +
                                                         "\n    Vorname:  " + old.vorname +
                                                         "\n    Nachname: " + old.nachname +
                                                         "\n    Firma:    " + old.rechnungsadresse_firma +
                                                         "\n  CSV Daten:" +
                                                         "\n    ID:       " + str(objid) +
                                                         "\n    Vorname:  " + vorname +
                                                         "\n    Nachname: " + nachname +
                                                         "\n    Firma:    " + firma
                                                         ))
                    if input("ID aus der CSV-Datei verwenden? [y/n] ").lower() in ["y", "yes", "j", "ja", "aktualisieren"]:
                        oldid = old.pk
                        old.pk = int(objid)
                        old.save()
                        new = old
                        old = Kunde.objects.get(objid=oldid)
                        new.zusammenfuegen = old
                        new.save()
                        self.stdout.write(
                            self.style.SUCCESS("ID aktualisiert!"))
                elif not oldwithemail and oldwithid:
                    old = Kunde.objects.get(pk=objid)
                    self.stdout.write(self.style.SUCCESS("Kunde mit der ID '"+str(objid)+"' existiert bereits, jedoch ist die E-Mail unterschiedlich!" +
                                                         "\n  KMUHelper-Daten:" +
                                                         "\n    E-Mail:   " + old.email +
                                                         "\n    Vorname:  " + old.vorname +
                                                         "\n    Nachname: " + old.nachname +
                                                         "\n    Firma:    " + old.rechnungsadresse_firma +
                                                         "\n  CSV Daten:" +
                                                         "\n    E-Mail:   " + email +
                                                         "\n    Vorname:  " + vorname +
                                                         "\n    Nachname: " + nachname +
                                                         "\n    Firma:    " + firma
                                                         ))
                    if input("E-Mail aus der CSV-Datei verwenden? [y/n] ").lower() in ["y", "yes", "j", "ja", "aktualisieren"]:
                        old.email = email
                        old.save()
                        self.stdout.write(
                            self.style.SUCCESS("E-Mail aktualisiert!"))
                elif oldwithemail and oldwithid:
                    self.stdout.write(self.style.SUCCESS(
                        "Irgendetwas stimmt nicht! Es besteht ein Kunde mit der E-Mail '"+email+"' aber ein Anderer mit der ID '"+str(objid)+"'."))

        except FileNotFoundError:
            raise CommandError('Datei nicht gefunden!')



# Generated by Django 3.2 on 2021-04-14 12:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0064_emailtemplate'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailtemplate',
            name='mail_to',
            field=models.CharField(blank=True, default='', max_length=50, verbose_name='Empfänger'),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-10 14:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0041_auto_20201010_1630'),
    ]

    operations = [
        migrations.AlterField(
            model_name='einstellung',
            name='id',
            field=models.CharField(max_length=50, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='einstellung',
            name='name',
            field=models.CharField(max_length=200, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='geheime_einstellung',
            name='id',
            field=models.CharField(max_length=50, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



from django.contrib import admin

from kmuhelper.modules.settings import models
from kmuhelper.overrides import CustomModelAdmin


@admin.register(models.Einstellung)
class EinstellungenAdmin(CustomModelAdmin):
    list_display = ('name', 'content_display')
    ordering = ('name',)

    search_fields = ['name', 'description']

    readonly_fields = ["id", "name", "description"]

    hidden = True

    def get_fields(self, request, obj=None):
        return ['description', f'content_{obj.typ}']

    # Permissions

    NO_ADD = True
    NO_DELETE = True


#

modeladmins = [
    (models.Einstellung, EinstellungenAdmin),
]



from django import forms

from kmuhelper import constants

class PDFOrderForm(forms.Form):
    preset = forms.ChoiceField(
        label="Vorlage",
        choices=(
            ('invoice', 'Rechnung'),
            ('delivery-note', 'Lieferschein'),
            ('payment-reminder', 'Zahlungserinnerung'),
        ),
        required=True,
        help_text="Die Vorlage definiert das Layout und die Inhalte der PDF-Datei",
    )
    title = forms.CharField(
        label="Titel",
        required=False,
        max_length=32,
        help_text="Z. B. 'Rechnung' oder 'Lieferschein' - Leer lassen für Standartwert der Vorlage",
    )
    text = forms.CharField(
        label="Text",
        required=False,
        widget=forms.Textarea,
        help_text="Dieser Text wird unterhalb des Titels angezeigt - Leer lassen für Standartwert der Vorlage",
    )
    language = forms.ChoiceField(
        label="Sprache",
        choices=constants.LANGUAGES,
        required=True,
        help_text="Die Sprache, in der die PDF-Datei generiert werden soll"
    )

    do_print = forms.BooleanField(
        label="Druckversion?",
        required=False,
        help_text="Wenn aktiviert, wird die PDF-Datei ohne Schnittmarker generiert"
    )
    do_download = forms.BooleanField(
        label="Herunterladen?",
        required=False,
        help_text="Datei automatisch heruntergeladen (dieses Verhalten kann je nach Browser variieren)"
    )

    def get_url_params(self):
        result = f"?custom&preset={self.cleaned_data['preset']}&language={self.cleaned_data['language']}"
        if self.cleaned_data['do_print']:
            result += '&print'
        if self.cleaned_data['do_download']:
            result += '&download'
        return result
        
    def update_order_settings(self, order):
        order.rechnungstitel = self.cleaned_data['title']
        order.rechnungstext = self.cleaned_data['text']
        order.save()



# Generated by Django 3.1.2 on 2020-10-18 12:19

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0044_auto_20201010_1746'),
    ]

    operations = [
        migrations.CreateModel(
            name='EMail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('typ', models.CharField(blank=True, choices=[('', '-'), ('kunde_registriert', 'Registrierungsmail Kunde')], default='', max_length=50, verbose_name='Typ')),
                ('subject', models.CharField(max_length=50, verbose_name='Betreff')),
                ('to', models.EmailField(max_length=254, verbose_name='Empfänger')),
                ('html_template', models.CharField(max_length=100, verbose_name='Name der Vorlage')),
                ('html_context', models.JSONField(verbose_name='Daten')),
                ('token', models.UUIDField(default=uuid.uuid4, editable=False, verbose_name='Token')),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Erstellt am')),
                ('time_sent', models.DateTimeField(blank=True, default=None, null=True, verbose_name='Gesendet um')),
            ],
            options={
                'verbose_name': 'E-Mail',
                'verbose_name_plural': 'E-Mails',
            },
        ),
        migrations.RemoveField(
            model_name='kunde',
            name='registrierungsemail_gesendet',
        ),
        migrations.AddField(
            model_name='kunde',
            name='registrierungsemail',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='kmuhelper.email'),
        ),
    ]



# Generated by Django 3.0.4 on 2020-07-16 13:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0020_auto_20200716_1443'),
    ]

    operations = [
        migrations.CreateModel(
            name='versandliste',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('kmuhelper.bestellung',),
        ),
    ]



# Generated by Django 3.0.4 on 2020-07-20 17:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0028_auto_20200720_1849'),
    ]

    operations = [
        migrations.AddField(
            model_name='notiz',
            name='produkt',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notiz', to='kmuhelper.Produkt'),
        ),
    ]



from django import template
from django.urls import reverse
from django.apps import apps

from kmuhelper import settings
from kmuhelper.modules.integrations.woocommerce.utils import is_connected as is_woocommerce_connected

register = template.Library()

##########


@register.simple_tag(takes_context=True, name="kmuhelper_woocommerce_connected")
def kmuhelper_woocommerce_connected(context=None):
    return is_woocommerce_connected()


@register.simple_tag(takes_context=True, name="kmuhelper_email_show_buttons")
def kmuhelper_email_show_buttons(context=None):
    show = bool(settings.get_db_setting("email-show-buttons"))
    return show


@register.filter
def kmuhelper_breadcrumbs(value, args=None):
    url_admin_kmuhelper = reverse(
        'admin:app_list', kwargs={'app_label': 'kmuhelper'})

    if args is None:
        title = "KMUHelper Admin"
        url = url_admin_kmuhelper
    else:
        title, viewname = args.split('|')
        url = reverse(viewname)

    value = "</a>".join(value.split(f'<a href="{url_admin_kmuhelper}">')[
                        1].split("</a>")[1::])
    value = f'<div id="breadcrumbs" class="breadcrumbs"><a href="{url}">{title}</a>' + value
    return value


@register.inclusion_tag('kmuhelper/_includes/pagechooser.html', takes_context=False)
def kmuhelper_pagechooser_from_model_list(model_list):
    griddata = []
    for model in model_list:
        try:
            dbmodel = apps.get_model('kmuhelper', model["object_name"])
        except LookupError:
            dbmodel = None
        griddata.append({
            "title": getattr(dbmodel, "admin_title", "") or model["name"],
            "url": model["admin_url"],
            "subtitle": getattr(dbmodel, "admin_description", ""),
            "icon": getattr(dbmodel, "admin_icon", "fa-solid fa-circle-exclamation"),
        })
    return {"griddata": [griddata], "pagechooserclass": "smallboxes"}


@register.inclusion_tag('kmuhelper/_includes/pagechooser.html', takes_context=False)
def kmuhelper_pagechooser(griddata, pagechooserclass=""):
    return {"griddata": griddata, "pagechooserclass": pagechooserclass}



# Generated by Django 3.2 on 2021-04-20 19:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0070_auto_20210417_2146'),
    ]

    operations = [
        migrations.CreateModel(
            name='PaymentImport',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time_imported', models.DateTimeField(auto_now_add=True, verbose_name='Importiert am')),
                ('is_parsed', models.BooleanField(default=False, verbose_name='Verarbeitet?')),
                ('data_msgid', models.CharField(max_length=50, verbose_name='Nachrichtenid')),
                ('data_creationdate', models.DateTimeField(verbose_name='Erstellt am')),
            ],
            options={
                'verbose_name': 'Zahlungsimport',
                'verbose_name_plural': 'Zahlungsimporte',
            },
        ),
        migrations.CreateModel(
            name='PaymentImportEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ref', models.CharField(max_length=50, verbose_name='Referenznummer')),
                ('iban', models.CharField(max_length=22, verbose_name='IBAN')),
                ('name', models.CharField(max_length=100, verbose_name='Name')),
                ('amount', models.FloatField(verbose_name='Betrag')),
                ('currency', models.CharField(max_length=10, verbose_name='Währung')),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='kmuhelper.paymentimport')),
            ],
            options={
                'verbose_name': 'Zahlungsimport-Einträge',
                'default_permissions': (),
            },
        ),
    ]



from django.urls import path, re_path

from kmuhelper.modules.api import views

#######################

urlpatterns = [
    path('versions', views.versions, name="api-versions"),
    path('orders/unpaid', views.orders_unpaid, name="api-orders-unpaid"),
    path('orders/<int:object_id>/set-paid', views.orders_set_paid, name="api-orders-set-paid"),

    re_path('.*', views.not_found, name="api-not-found")
]



# Generated by Django 3.2 on 2021-04-23 15:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0073_auto_20210422_1323'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bestellung',
            name='status',
            field=models.CharField(choices=[('pending', 'Zahlung ausstehend'), ('processing', 'In Bearbeitung'), ('on-hold', 'In Wartestellung'), ('completed', 'Abgeschlossen'), ('cancelled', 'Storniert/Abgebrochen'), ('refunded', 'Rückerstattet'), ('failed', 'Fehlgeschlagen'), ('trash', 'Gelöscht')], default='processing', max_length=11, verbose_name='Status'),
        ),
    ]



# Generated by Django 3.1 on 2020-09-22 19:49

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0033_auto_20200802_2105'),
    ]

    operations = [
        migrations.AddField(
            model_name='bestellungsposten',
            name='rabatt',
            field=models.IntegerField(default=0, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)], verbose_name='Rabatt in %'),
        ),
    ]



# Generated by Django 3.2 on 2021-04-17 18:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0068_rename_registrierungsemail_kunde_email_registriert'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='emailattachment',
            options={'verbose_name': 'E-Mail-Anhang', 'verbose_name_plural': 'E-Mail-Anhänge'},
        ),
        migrations.AlterModelOptions(
            name='emailtemplate',
            options={'verbose_name': 'E-Mail-Vorlage', 'verbose_name_plural': 'E-Mail-Vorlagen'},
        ),
        migrations.AlterField(
            model_name='bestellung',
            name='rechnungsadresse_adresszeile1',
            field=models.CharField(blank=True, default='', help_text="Strasse und Hausnummer oder 'Postfach'", max_length=50, verbose_name='Adresszeile 1'),
        ),
    ]



# Generated by Django 3.2 on 2021-04-13 21:32

from django.db import migrations, models
import kmuhelper.external.multi_email_field.fields


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0062_email_text'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='bcc',
            field=kmuhelper.external.multi_email_field.fields.MultiEmailField(blank=True, default='', help_text='Blindkopie', verbose_name='BCC'),
        ),
        migrations.AlterField(
            model_name='email',
            name='cc',
            field=kmuhelper.external.multi_email_field.fields.MultiEmailField(blank=True, default='', help_text='Kopie', verbose_name='CC'),
        ),
        migrations.AlterField(
            model_name='email',
            name='html_context',
            field=models.JSONField(blank=True, default=dict, help_text='Daten im JSON-Format, mit welchen die Designvorlage befüllt wird.', null=True, verbose_name='Daten'),
        ),
        migrations.AlterField(
            model_name='email',
            name='html_template',
            field=models.CharField(default='default.html', help_text="Dateiname der Designvorlage unter 'kmuhelper/emails/'.", max_length=50, verbose_name='Designvorlage'),
        ),
        migrations.AlterField(
            model_name='email',
            name='notes',
            field=models.TextField(blank=True, default='', help_text='Diese Notizen haben keine Einwirkung auf die E-Mail selbst.', verbose_name='Notizen'),
        ),
        migrations.AlterField(
            model_name='email',
            name='subject',
            field=models.CharField(help_text='Wird Standardmässig auch als Untertitel verwendet.', max_length=50, verbose_name='Betreff'),
        ),
        migrations.AlterField(
            model_name='email',
            name='text',
            field=models.TextField(blank=True, default='', help_text='Hauptinhalt - wird nicht von allen Designvorlagen verwendet. Links und E-Mail Adressen werden automatisch verlinkt.', verbose_name='Text'),
        ),
        migrations.AlterField(
            model_name='email',
            name='time_created',
            field=models.DateTimeField(auto_now_add=True, help_text='Datum und Zeit der Erstellung dieser E-Mail.', verbose_name='Erstellt am'),
        ),
        migrations.AlterField(
            model_name='email',
            name='time_sent',
            field=models.DateTimeField(blank=True, default=None, help_text='Datum und Zeit des letzten erfolgreichen Sendeversuches.', null=True, verbose_name='Gesendet um'),
        ),
        migrations.AlterField(
            model_name='email',
            name='to',
            field=kmuhelper.external.multi_email_field.fields.MultiEmailField(default=[], help_text='Direkte Empfänger', verbose_name='Empfänger'),
        ),
    ]



from django.urls import path

from kmuhelper.modules.app import views

#######################

urlpatterns = [
    path('mobile/', views.app_mobile_main, name="app-mobile"),
    path('mobile/home', views.app_mobile_home, name="app-mobile-home"),
    path('mobile/error', views.app_mobile_error, name="app-mobile-error"),

    path('mobile/manifest.webmanifest', views.app_mobile_manifest, name="app-mobile-manifest"),

    path('desktop/', views.app_desktop, name="app-desktop"),
]



import sys
import subprocess
import requests

from io import BytesIO
from packaging.version import Version, InvalidVersion

from django.contrib import messages
from django.core import mail
from django.shortcuts import render
from django.template.loader import get_template
from django.urls import reverse

from kmuhelper.constants import URL_FAQ

################

def render_error(request, status:int=404, message:str=""):
    """Show the error page"""

    if message:
        messages.error(request, message)

    return render(request, "kmuhelper/error.html", status=status)

################

def python_version():
    return str(sys.version.split(" ")[0])


def _package_versions_pypi(project, testpypi=False, allow_prereleases=False):
    url = "https://"+("test." if testpypi else "") + \
        "pypi.org/pypi/"+project+"/json"
    versions = requests.get(url).json()["releases"].keys()
    versions_safe = []
    for v in versions:
        try:
            vs = Version(v)
            if not allow_prereleases and vs.is_prerelease:
                continue
            versions_safe.append(vs)
        except (InvalidVersion):
            continue
    return sorted(versions_safe)

def _package_version_local(package) -> Version:
    cmd = [sys.executable, '-m', 'pip', 'show', '{}'.format(package)]
    ver = str(subprocess.run(cmd, capture_output=True, text=True))
    ver = ver[ver.find('Version:')+8:]
    ver = ver[:ver.find('\\n')].replace(' ', '')
    try:
        return Version(ver)
    except (InvalidVersion):
        return None

def package_version(package, testpypi=False):
    all_versions = _package_versions_pypi(package, testpypi)
    latest_version = all_versions[-1] if all_versions else None
    current_version = _package_version_local(package)
    
    if latest_version and current_version:
        uptodate = latest_version <= current_version
    else:
        uptodate = None

    return {
        # "all":      all_versions,
        "latest":    str(latest_version),
        "current":   str(current_version),
        "uptodate":  uptodate,
    }

################


def getfirstindex(data: list, search: list):
    """Get the index of the first occurence of any string from search in data"""
    for s in search:
        if s in data:
            return data.index(s)
    return None


def formatprice(preis):
    """Format a float to exactly 2 decimal places"""
    return "{:.2f}".format(float(preis))


def runden(preis, to=0.05):
    """Round a float to .05 or any other value"""
    # Note: formatprice is used because of floating point approximation
    return float(formatprice(float(round(round(preis / to) * to, 2))))


###############


def send_mail(subject: str, to: str, template_name: str, context: dict = {}, **kwargs):
    html_message = get_template(
        "kmuhelper/emails/"+template_name).render(context)

    msg = mail.EmailMessage(
        subject=subject,
        body=html_message,
        to=[to],
        **kwargs
    )

    msg.content_subtype = "html"

    return bool(msg.send())


def send_pdf(subject: str, to: str, template_name: str, pdf: BytesIO, pdf_filename: str = "file.pdf", context: dict = {}, **kwargs):
    html_message = get_template(
        "kmuhelper/emails/"+template_name).render(context)

    msg = mail.EmailMessage(
        subject=subject,
        body=html_message,
        to=[to],
        **kwargs
    )

    msg.content_subtype = "html"
    msg.attach(filename=pdf_filename, content=pdf.read(),
               mimetype="application/pdf")

    return bool(msg.send())

###############


def modulo10rekursiv(strNummer):
    intTabelle = [
        [0, 9, 4, 6, 8, 2, 7, 1, 3, 5],
        [9, 4, 6, 8, 2, 7, 1, 3, 5, 0],
        [4, 6, 8, 2, 7, 1, 3, 5, 0, 9],
        [6, 8, 2, 7, 1, 3, 5, 0, 9, 4],
        [8, 2, 7, 1, 3, 5, 0, 9, 4, 6],
        [2, 7, 1, 3, 5, 0, 9, 4, 6, 8],
        [7, 1, 3, 5, 0, 9, 4, 6, 8, 2],
        [1, 3, 5, 0, 9, 4, 6, 8, 2, 7],
        [3, 5, 0, 9, 4, 6, 8, 2, 7, 1],
        [5, 0, 9, 4, 6, 8, 2, 7, 1, 3],
    ]
    strNummer = strNummer.replace(" ", "")
    uebertrag = 0
    for num in strNummer:
        uebertrag = intTabelle[uebertrag][int(num)]
    return [0, 9, 8, 7, 6, 5, 4, 3, 2, 1][uebertrag]


###############

def custom_app_list(request, models, title, url):
    "Get a custom app_list for the admin site"
    return [{
        'app_label': 'kmuhelper',
        'app_url': url,
        'has_module_perms': True,
        'models': [
            {
                'add_url': reverse(f"admin:kmuhelper_{model._meta.model_name}_add"),
                'admin_url': reverse(f"admin:kmuhelper_{model._meta.model_name}_changelist"),
                'name': model._meta.verbose_name_plural,
                'object_name': model._meta.model_name,
                'perms': {
                    'add': request.user.has_perm(f"kmuhelper.add_{model._meta.model_name}"),
                    'change': request.user.has_perm(f"kmuhelper.change_{model._meta.model_name}"),
                    'delete': request.user.has_perm(f"kmuhelper.delete_{model._meta.model_name}"),
                    'view': request.user.has_perm(f"kmuhelper.view_{model._meta.model_name}")},
                'view_only': False
            } for model in models
        ],
        'name': title,
    }]

###############

def faq(id="", text="FAQ"):
    "Get a link to the FAQ page in the manual."
    return f'<a href="{URL_FAQ}#{id}" target="_blank">{text}</a>'



# Generated by Django 3.0.4 on 2020-04-26 16:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0002_auto_20200426_1746'),
    ]

    operations = [
        migrations.AlterField(
            model_name='kosten',
            name='name',
            field=models.CharField(default='Zusätzliche Kosten', max_length=500, verbose_name='Name'),
        ),
    ]



# Generated by Django 3.2.10 on 2022-03-11 20:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0079_auto_20220305_2322'),
    ]

    operations = [
        migrations.DeleteModel(
            name='ToDoLagerbestand',
        ),
        migrations.DeleteModel(
            name='ToDoLieferung',
        ),
        migrations.DeleteModel(
            name='ToDoNotiz',
        ),
        migrations.DeleteModel(
            name='ToDoVersand',
        ),
        migrations.DeleteModel(
            name='ToDoZahlungseingang',
        ),
        migrations.CreateModel(
            name='App_Lagerbestand',
            fields=[
            ],
            options={
                'verbose_name': 'Produkt',
                'verbose_name_plural': 'Produkte',
                'proxy': True,
                'default_permissions': ('add', 'change', 'view'),
                'indexes': [],
                'constraints': [],
            },
            bases=('kmuhelper.produkt',),
        ),
        migrations.CreateModel(
            name='App_ToDo',
            fields=[
            ],
            options={
                'verbose_name': 'Notiz',
                'verbose_name_plural': 'Notizen',
                'proxy': True,
                'default_permissions': ('add', 'change', 'view'),
                'indexes': [],
                'constraints': [],
            },
            bases=('kmuhelper.notiz',),
        ),
        migrations.CreateModel(
            name='App_Warenausgang',
            fields=[
            ],
            options={
                'verbose_name': 'Bestellung',
                'verbose_name_plural': 'Bestellungen',
                'proxy': True,
                'default_permissions': ('add', 'change', 'view'),
                'indexes': [],
                'constraints': [],
            },
            bases=('kmuhelper.bestellung',),
        ),
        migrations.CreateModel(
            name='App_Wareneingang',
            fields=[
            ],
            options={
                'verbose_name': 'Lieferung',
                'verbose_name_plural': 'Lieferungen',
                'proxy': True,
                'default_permissions': ('add', 'change', 'view'),
                'indexes': [],
                'constraints': [],
            },
            bases=('kmuhelper.lieferung',),
        ),
        migrations.CreateModel(
            name='App_Zahlungseingang',
            fields=[
            ],
            options={
                'verbose_name': 'Bestellung',
                'verbose_name_plural': 'Bestellung',
                'proxy': True,
                'default_permissions': ('add', 'change', 'view'),
                'indexes': [],
                'constraints': [],
            },
            bases=('kmuhelper.bestellung',),
        ),
    ]



# Generated by Django 4.1.5 on 2023-01-29 19:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("kmuhelper", "0084_alter_produkt_mengenbezeichnung"),
    ]

    operations = [
        migrations.AlterField(
            model_name="bestellungskosten",
            name="name",
            field=models.CharField(
                default="Zusätzliche Kosten",
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                max_length=500,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="kosten",
            name="name",
            field=models.CharField(
                default="Zusätzliche Kosten",
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                max_length=500,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="produkt",
            name="beschrieb",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                verbose_name="Beschrieb",
            ),
        ),
        migrations.AlterField(
            model_name="produkt",
            name="kurzbeschrieb",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                verbose_name="Kurzbeschrieb",
            ),
        ),
        migrations.AlterField(
            model_name="produkt",
            name="mengenbezeichnung",
            field=models.CharField(
                blank=True,
                default="Stück",
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                max_length=100,
                verbose_name="Mengenbezeichnung",
            ),
        ),
        migrations.AlterField(
            model_name="produkt",
            name="name",
            field=models.CharField(
                help_text="Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'",
                max_length=500,
                verbose_name="Name",
            ),
        ),
    ]



from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy

from kmuhelper.decorators import require_object, confirm_action
from kmuhelper.modules.emails.models import EMail, Attachment, EMailTemplate
from kmuhelper.utils import render_error, custom_app_list

#####

# EMail views


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required("kmuhelper.send_email")
@require_object(EMail)
@confirm_action("E-Mail Nachricht senden")
def email_send(request, obj):
    if obj.is_valid():
        try:
            success = obj.send()
            if success:
                messages.success(request, "E-Mail wurde versendet!")
            else:
                messages.error(request, "E-Mail konnte nicht gesendet werden!")
        except FileNotFoundError:
            messages.error(request, "Eine Datei aus dem Anhang konnte nicht gefunden werden!")
    return redirect(reverse("admin:kmuhelper_email_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required("kmuhelper.send_email")
@require_object(EMail)
@confirm_action("E-Mail Nachricht ERNEUT senden")
def email_resend(request, obj):
    if obj.is_valid(request):
        try:
            success = obj.send()
            if success:
                messages.success(request, "E-Mail wurde erneut versendet!")
            else:
                messages.error(request, "E-Mail konnte nicht gesendet werden!")
        except FileNotFoundError:
            messages.error(request, "Eine Datei aus dem Anhang konnte nicht gefunden werden!")
    return redirect(reverse("admin:kmuhelper_email_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required("kmuhelper.view_email")
@require_object(EMail)
def email_preview(request, obj):
    if obj.is_valid(request):
        return HttpResponse(obj.render(online=True))
    return redirect(reverse("admin:kmuhelper_email_change", args=[obj.pk]))


# Public views


@require_object(EMail, show_errorpage=True)
def email_view(request, obj):
    """Render an email for online viewing"""

    token_received = request.GET.get("token", None)
    token_stored = str(obj.token)

    if request.user.has_perm("kmuhelper.view_email") or (token_received == token_stored):
        if obj.is_valid():
            return HttpResponse(obj.render(online=True))

        messages.warning(
            request, "Diese E-Mail kann zurzeit leider nicht angezeigt werden.")
    else:
        messages.error(request, "Du hast keinen Zugriff auf diese E-Mail!")

    return render_error(request)


# EMailAttachment views


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required("kmuhelper.download_attachment")
@require_object(Attachment)
def attachment_download(request, obj):
    download = not "preview" in request.GET
    try:
        return obj.get_file_response(download=download)
    except FileNotFoundError:
        messages.error(request, "Datei wurde nicht gefunden!")
        return render_error(request)


# Public views


@require_object(Attachment, show_errorpage=True)
def attachment_view(request, obj):
    """Render an attachment for online viewing"""

    token_received = request.GET.get("token", None)
    token_stored = str(obj.token)

    if request.user.has_perm("kmuhelper.download_attachment") or (token_received == token_stored):
        download = "download" in request.GET
        try:
            return obj.get_file_response(download=download)
        except FileNotFoundError:
            messages.error(request, "Diese Datei ist leider nicht mehr verfügbar!")
    else:
        messages.error(request, "Du hast keinen Zugriff auf diesen E-Mail-Anhang!")

    return render_error(request)


# EMailTemplate Views


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required(["kmuhelper.view_emailtemplate"])
def emailtemplate_savevars(request):
    """Save variables into session for usage in template"""

    data = request.GET.dict()
    request.session["emailtemplate-vars"] = data

    messages.success(
        request, f"Folgende Daten wurden gespeichert, um die nächste Vorlage auszufüllen: {data}")
    return redirect(reverse("admin:kmuhelper_emailtemplate_changelist"))


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required(["kmuhelper.view_emailtemplate"])
def emailtemplate_resetvars(request):
    """Reset saved session variables for email templates"""

    if "emailtemplate-vars" in request.session:
        del request.session["emailtemplate-vars"]

    messages.success(
        request, "Gespeicherte Daten wurden gelöscht!")
    return redirect(reverse("admin:kmuhelper_emailtemplate_changelist"))


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required(["kmuhelper.view_emailtemplate", "kmuhelper.add_email"])
@require_object(EMailTemplate)
def emailtemplate_use(request, obj):
    """Use a template"""

    savedvars = request.session.pop("emailtemplate-vars", dict())
    requestvars = request.GET.dict()
    data = {**savedvars, **requestvars}

    mail = obj.create_mail(data)

    messages.success(
        request, f"Vorlage wurde mit folgenden Daten ausgefüllt: {data}")
    return redirect(reverse("admin:kmuhelper_email_change", args=[mail.pk]))


# Custom fake admin

@login_required(login_url=reverse_lazy("admin:login"))
def email_index(request):
    return render(request, 'admin/kmuhelper/_special/emails/app_index.html', {
        'app_label': 'kmuhelper',
        'app_list': custom_app_list(request, [EMail, EMailTemplate, Attachment], "KMUHelper E-Mails", reverse('kmuhelper:email-index')),
        'has_permission': True,
        'is_nav_sidebar_enabled': False,
        'is_popup': False,
        'title': 'KMUHelper E-Mails',
    })
